# Copyright 2014, 2015, 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import errno
import fixtures
import gc
from lxml import etree
import six
import subunit

if six.PY2:
    import __builtin__ as builtins
elif six.PY3:
    import builtins

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

import mock
import requests.models as req_mod
import requests.structures as req_struct
import testtools

import pypowervm.adapter as adp
import pypowervm.const as c
import pypowervm.entities as ent
import pypowervm.exceptions as pvmex
import pypowervm.tests.lib as testlib
import pypowervm.tests.test_fixtures as fx
from pypowervm.tests.test_utils import pvmhttp
from pypowervm.wrappers import storage as pvm_stor

logon_text = testlib.file2b("logon.xml")
response_text = testlib.file2b("event.xml")
NET_BRIDGE_FILE = 'fake_network_bridge.txt'


class TestAdapter(testtools.TestCase):
    """Test cases to test the adapter classes and methods."""

    def _mk_response(self, status, content=None):
        reasons = {200: 'OK', 204: 'No Content', 401: 'Unauthorized'}
        # Create a Response object, that will serve as a mock return value
        my_response = req_mod.Response()
        my_response.status_code = status
        my_response.reason = reasons[status]
        clen = '0'
        if status == 200 and content:
            clen = str(len(content))
        dict_headers = {
            'content-length': clen, 'x-powered-by': 'Servlet/3.0',
            'set-cookie': ('JSESSIONID=0000a41BnJsGTNQvBGERA3wR1nj:759878cb-4f'
                           '9a-4b05-a09a-3357abfea3b4; Path=/; Secure; HttpOnl'
                           'y, CCFWSESSION=E4C0FFBE9130431DBF1864171ECC6A6E; P'
                           'ath=/; Secure; HttpOnly'),
            'expires': 'Thu, 01 Dec 1994 16:00:00 GMT',
            'x-transaction-id': 'XT10000073',
            'cache-control': 'no-cache="set-cookie, set-cookie2"',
            'date': 'Wed, 23 Jul 2014 21:51:10 GMT',
            'content-type': 'application/vnd.ibm.powervm'}
        my_response.headers = req_struct.CaseInsensitiveDict(dict_headers)
        my_response._content = content
        return my_response

    def setUp(self):
        super(TestAdapter, self).setUp()
        """Set up a mocked Session instance."""
        # Init test data
        host = '0.0.0.0'
        user = 'user'
        pwd = 'pwd'
        auditmemento = 'audit'

        # Create a Response object, that will serve as a mock return value
        my_response = self._mk_response(200, logon_text)

        # Mock out the method and class we are not currently testing
        with mock.patch('requests.Session') as mock_session:
            session = mock_session.return_value
            session.request.return_value = my_response

            # Create session for the test to use
            self.sess = adp.Session(host, user, pwd,
                                    auditmemento=auditmemento,
                                    certpath=None)
            # Mock out the logoff, which gets called when the session
            # goes out of scope during tearDown()
            self.sess._logoff = mock.Mock()

    def tearDown(self):
        """Tear down the Session instance."""
        self.sess = None
        super(TestAdapter, self).tearDown()

    @mock.patch('pypowervm.wrappers.event.Event.wrap')
    @mock.patch('time.sleep')
    def test_event_listener(self, mock_sleep, mock_evt_wrap):

        with mock.patch.object(adp._EventListener, '_get_events') as m_events,\
                mock.patch.object(adp, '_EventPollThread') as mock_poll:
            # With some fake events, event listener can be initialized
            self.sess._sessToken = 'token'.encode('utf-8')
            m_events.return_value = {'general': 'init'}, 'raw_evt', 'wrap_evt'
            event_listen = self.sess.get_event_listener()
            self.assertIsNotNone(event_listen)

            # Register the fake handlers and ensure they are called
            evh = mock.Mock(spec=adp.EventHandler, autospec=True)
            raw_evh = mock.Mock(spec=adp.RawEventHandler, autospec=True)
            wrap_evh = mock.Mock(spec=adp.WrapperEventHandler, autospec=True)
            event_listen.subscribe(evh)
            event_listen.subscribe(raw_evh)
            event_listen.subscribe(wrap_evh)
            events, raw_events, evtwraps = event_listen._get_events()
            event_listen._dispatch_events(events, raw_events, evtwraps)
            evh.process.assert_called_once_with({'general': 'init'})
            raw_evh.process.assert_called_once_with('raw_evt')
            wrap_evh.process.assert_called_once_with('wrap_evt')
            self.assertTrue(mock_poll.return_value.start.called)

            # Ensure getevents() gets legacy events
            self.assertEqual({'general': 'init'}, event_listen.getevents())

        # Outside our patching of _get_events, get the formatted events
        with mock.patch.object(event_listen, '_format_events') as mock_format,\
                mock.patch.object(event_listen.adp, 'read') as mock_read:

            # Ensure exception path doesn't kill the thread
            mock_read.side_effect = Exception()
            self.assertEqual(({}, [], []), event_listen._get_events())
            self.assertEqual(1, mock_read.call_count)
            mock_format.assert_not_called()
            mock_evt_wrap.assert_not_called()
            mock_sleep.assert_called_once_with(5)

            mock_read.reset_mock()
            # side_effect takes precedence over return_value; so kill it.
            mock_read.side_effect = None

            # Fabricate some mock entries, so format gets called.
            mock_read.return_value.feed.entries = (['entry1', 'entry2'])

            self.assertEqual(({}, [], mock_evt_wrap.return_value),
                             event_listen._get_events())
            self.assertEqual(1, mock_read.call_count)
            mock_format.assert_has_calls([mock.call('entry1', {}, []),
                                          mock.call('entry2', {}, [])])
            mock_evt_wrap.assert_called_once_with(mock_read.return_value)

        # Test _format_events
        event_data = [
            {
                'EventType': 'NEW_CLIENT',
                'EventData': 'href1',
                'EventID': '1',
                'EventDetail': 'detail',
            },
            {
                'EventType': 'CACHE_CLEARED',
                'EventData': 'href2',
                'EventID': '2',
                'EventDetail': 'detail2',
            },
            {
                'EventType': 'ADD_URI',
                'EventData': 'LPAR1',
                'EventID': '3',
                'EventDetail': 'detail3',
            },
            {
                'EventType': 'DELETE_URI',
                'EventData': 'LPAR1',
                'EventID': '4',
                'EventDetail': 'detail4',
            },
            {
                'EventType': 'INVALID_URI',
                'EventData': 'LPAR1',
                'EventID': '4',
                'EventDetail': 'detail4',
            },
        ]

        # Setup a side effect that returns events from the test data.
        def get_event_data(item):
            data = event_data[0][item]
            if item == 'EventDetail':
                event_data.pop(0)
            return data

        # Raw events returns a sequence the same as the test data
        raw_result = copy.deepcopy(event_data)
        # Legacy events overwrites some events.
        dict_result = {'general': 'invalidate', 'LPAR1': 'delete'}

        # Build a mock entry
        entry = mock.Mock()
        entry.element.findtext.side_effect = get_event_data
        events = {}
        raw_events = []
        x = len(raw_result)
        while x:
            x -= 1
            event_listen._format_events(entry, events, raw_events)
        self.assertEqual(raw_result, raw_events)
        self.assertEqual(dict_result, events)

    @mock.patch('pypowervm.adapter.Session')
    def test_empty_init(self, mock_sess):
        adp.Adapter()
        mock_sess.assert_called_with()

    def test_no_cache(self):
        self.assertRaises(pvmex.CacheNotSupportedException,
                          adp.Adapter, use_cache=True)

    @mock.patch('requests.Session')
    def test_read(self, mock_session):
        """Test read() method found in the Adapter class."""
        # Init test data
        root_type = 'ManagedSystem'
        root_id = 'caae9209-25e5-35cd-a71a-ed55c03f294d'
        child_type = 'child'
        child_id = 'child'
        suffix_type = 'quick'
        adapter = adp.Adapter(self.sess)

        # Create a Response object, that will serve as a mock return value
        read_response = self._mk_response(200, response_text)

        # Mock out the method and class we are not currently testing
        session = mock_session.return_value
        session.request.return_value = read_response

        # Run the actual test
        ret_read_value = adapter.read(root_type, root_id, child_type,
                                      child_id, suffix_type)

        # Verify Correct path was built in build_path()
        reqpath = adp.Adapter.build_path('uom', root_type, root_id,
                                         child_type, child_id, suffix_type)
        # Verify the return value
        # self.assertIsInstance(ret_read_value, adp.Response)
        self.assertEqual('GET', ret_read_value.reqmethod)
        self.assertEqual(200, ret_read_value.status)
        self.assertEqual(reqpath, ret_read_value.reqpath)

    @mock.patch('pypowervm.adapter.Adapter._validate')
    @mock.patch('pypowervm.adapter.Adapter.build_path')
    @mock.patch('pypowervm.adapter.Adapter.read_by_path')
    def test_read2(self, mock_rbp, mock_bld, mock_val):
        """Validate shallow flow & arg passing."""
        adap = adp.Adapter(session=self.sess)
        # Defaults
        self.assertEqual(mock_rbp.return_value, adap.read('root_type'))
        mock_val.assert_called_once_with(
            'read', 'root_type', None, None, None, None, None, None)
        mock_bld.assert_called_once_with(
            'uom', 'root_type', None, None, None, None, None, None, xag=None,
            add_qp=None)
        mock_rbp.assert_called_once_with(
            mock_bld.return_value, None, timeout=-1, auditmemento=None, age=-1,
            sensitive=False, helpers=None)
        # Specified kwargs
        mock_val.reset_mock()
        mock_bld.reset_mock()
        mock_rbp.reset_mock()
        self.assertEqual(mock_rbp.return_value, adap.read(
            'root_type', root_id='root_id', child_type='child_type',
            child_id='child_id', suffix_type='suffix_type',
            suffix_parm='suffix_parm', detail='detail', service='service',
            etag='etag', timeout='timeout', auditmemento='auditmemento',
            age='age', xag='xag', sensitive='sensitive', helpers='helpers',
            add_qp='add_qp'))
        mock_val.assert_called_once_with(
            'read', 'root_type', 'root_id', 'child_type', 'child_id',
            'suffix_type', 'suffix_parm', 'detail')
        mock_bld.assert_called_once_with(
            'service', 'root_type', 'root_id', 'child_type', 'child_id',
            'suffix_type', 'suffix_parm', 'detail', xag='xag', add_qp='add_qp')
        mock_rbp.assert_called_once_with(
            mock_bld.return_value, 'etag', timeout='timeout',
            auditmemento='auditmemento', age='age', sensitive='sensitive',
            helpers='helpers')

    @mock.patch('pypowervm.adapter.Adapter.extend_path')
    def test_build_path(self, mock_exp):
        """Validate build_path."""
        adap = adp.Adapter(session=self.sess)
        # Defaults
        self.assertEqual(mock_exp.return_value, adap.build_path(
            'service', 'root_type'))
        mock_exp.assert_called_once_with(
            '/rest/api/service/root_type', suffix_type=None, suffix_parm=None,
            detail=None, xag=None, add_qp=None)
        # child specs ignored if no root ID
        mock_exp.reset_mock()
        self.assertEqual(mock_exp.return_value, adap.build_path(
            'service', 'root_type', child_type='child_type',
            child_id='child_id'))
        mock_exp.assert_called_once_with(
            '/rest/api/service/root_type', suffix_type=None, suffix_parm=None,
            detail=None, xag=None, add_qp=None)
        # child ID ignored if no child type
        mock_exp.reset_mock()
        self.assertEqual(mock_exp.return_value, adap.build_path(
            'service', 'root_type', root_id='root_id', child_id='child_id'))
        mock_exp.assert_called_once_with(
            '/rest/api/service/root_type/root_id', suffix_type=None,
            suffix_parm=None, detail=None, xag=None, add_qp=None)
        # Specified kwargs (including full child spec
        mock_exp.reset_mock()
        self.assertEqual(mock_exp.return_value, adap.build_path(
            'service', 'root_type', root_id='root_id', child_type='child_type',
            child_id='child_id', suffix_type='suffix_type',
            suffix_parm='suffix_parm', detail='detail', xag='xag',
            add_qp='add_qp'))
        mock_exp.assert_called_once_with(
            '/rest/api/service/root_type/root_id/child_type/child_id',
            suffix_type='suffix_type', suffix_parm='suffix_parm',
            detail='detail', xag='xag', add_qp='add_qp')

    @mock.patch('pypowervm.adapter.Adapter._request')
    def test_headers(self, mock_request):
        def validate_hdrs_func(acc=None, inm=None):
            expected_headers = {}
            if acc is not None:
                expected_headers['Accept'] = acc
            if inm is not None:
                expected_headers['If-None-Match'] = inm

            def validate_request(meth, path, **kwargs):
                self.assertEqual(expected_headers, kwargs['headers'])

            return validate_request

        adpt = adp.Adapter(mock.Mock())
        basepath = c.API_BASE_PATH + 'uom/SomeRootObject'
        uuid = "abcdef01-2345-2345-2345-67890abcdef0"
        hdr_xml = 'application/atom+xml'
        hdr_json = '*/*'
        etag = 'abc123'

        # Root feed
        mock_request.side_effect = validate_hdrs_func(acc=hdr_xml)
        adpt._read_by_path(basepath, None, None, None, None)

        # Root instance with etag
        mock_request.side_effect = validate_hdrs_func(acc=hdr_xml, inm=etag)
        adpt._read_by_path(basepath + '/' + uuid, etag, None, None, None)

        # Quick root anchor (produces XML report of available quick properties
        mock_request.side_effect = validate_hdrs_func(acc=hdr_xml)
        adpt._read_by_path(basepath + '/quick', None, None, None, None)

        # Quick root instance (JSON of all quick properties)
        mock_request.side_effect = validate_hdrs_func(acc=hdr_json)
        adpt._read_by_path('/'.join([basepath, uuid, 'quick']), None, None,
                           None, None)

        # Specific quick property
        mock_request.side_effect = validate_hdrs_func(acc=hdr_json)
        adpt._read_by_path('/'.join([basepath, uuid, 'quick', 'property']),
                           None, None, None, None)

        # Explicit JSON file
        mock_request.side_effect = validate_hdrs_func(acc=hdr_json)
        adpt._read_by_path('/'.join([basepath, 'somefile.json']), None, None,
                           None, None)

        # Object that happens to end in 'json'
        mock_request.side_effect = validate_hdrs_func(acc=hdr_xml)
        adpt._read_by_path('/'.join([basepath, 'xml_about_json']), None, None,
                           None, None)

        # Quick with query params and fragments
        mock_request.side_effect = validate_hdrs_func(acc=hdr_json)
        adpt._read_by_path('/'.join([basepath, uuid, 'quick']) +
                           '?group=None#frag', None, None, None, None)

    @mock.patch('requests.Session')
    def test_create(self, mock_session):
        """Test create() method found in the Adapter class."""
        # Init test data
        adapter = adp.Adapter(self.sess)
        new_scsi = pvm_stor.VSCSIClientAdapterElement.bld(adapter)

        element = new_scsi
        root_type = 'ManagedSystem'
        root_id = 'id'
        child_type = 'LogicalPartition'

        create_response = self._mk_response(200, response_text)

        # Mock out the method and class we are not currently testing
        session = mock_session.return_value
        session.request.return_value = create_response

        # Run the actual test
        ret_create_value = adapter.create(element, root_type, root_id,
                                          child_type)

        # Verify Correct path was built in build_path()
        reqpath = adp.Adapter.build_path('uom', root_type, root_id,
                                         child_type, xag=[])

        # Verify the return value
        # self.assertIsInstance(ret_create_value, adp.Response)
        self.assertEqual('PUT', ret_create_value.reqmethod)
        self.assertEqual(200, ret_create_value.status)
        self.assertEqual(reqpath, ret_create_value.reqpath)

    @mock.patch('requests.Session')
    def test_update(self, mock_session):
        """Test update() method found in the Adapter class."""
        # Init test data
        data = 'data'
        etag = 'etag'
        root_type = 'root type'
        root_id = 'root id'
        adapter = adp.Adapter(self.sess)

        update_response = self._mk_response(200, response_text)

        # Mock out the method and class we are not currently testing
        session = mock_session.return_value
        session.request.return_value = update_response

        # Run the actual test
        ret_update_value = adapter.update(data, etag, root_type, root_id)

        # Verify Correct path was built in build_path()
        reqpath = adp.Adapter.build_path('uom', root_type, root_id)

        # Verify the return value
        # self.assertIsInstance(ret_update_value, adp.Response)
        self.assertEqual('POST', ret_update_value.reqmethod)
        self.assertEqual(200, ret_update_value.status)
        self.assertEqual(reqpath, ret_update_value.reqpath)

    @mock.patch('requests.Session')
    def test_upload(self, mock_session):
        # Build the adapter
        adapter = adp.Adapter(self.sess)

        # Mock data
        filedesc_mock = mock.MagicMock()
        filedesc_mock.findtext.side_effect = ['uuid', 'mime']

        with mock.patch.object(adapter, '_request') as mock_request:
            adapter.upload_file(filedesc_mock, None)

        # Validate
        expected_headers = {'Accept': 'application/vnd.ibm.powervm.web+xml',
                            'Content-Type': 'mime'}
        expected_path = '/rest/api/web/File/contents/uuid'
        mock_request.assert_called_once_with(
            'PUT', expected_path, helpers=None, headers=expected_headers,
            timeout=-1, auditmemento=None, filehandle=None, chunksize=65536)

    def _test_upload_request(self, mock_rq, mock_fh, fhdata):
        """Test an upload requests with different kinds of "filehandle"."""
        adapter = adp.Adapter(self.sess)
        mock_fd = mock.Mock(findtext=mock.Mock(side_effect=['uuid', 'mime']))

        def check_request(method, url, data=None, headers=None, timeout=None):
            """Validate the session.request call."""
            self.assertEqual('PUT', method)
            self.assertEqual(
                self.sess.dest + '/rest/api/web/File/contents/uuid', url)
            # Verify that data is iterable
            self.assertEqual(fhdata, [chunk for chunk in data])
            return mock.Mock(status_code=c.HTTPStatus.OK_NO_CONTENT)
        mock_rq.side_effect = check_request

        adapter.upload_file(mock_fd, mock_fh)

    @mock.patch('requests.sessions.Session.request')
    def test_upload_request_iter(self, mock_rq):
        """Test an upload request with an iterable."""
        fhdata = ['one', 'two']
        self._test_upload_request(mock_rq, fhdata, fhdata)

    @mock.patch('requests.sessions.Session.request')
    def test_upload_request_fh(self, mock_rq):
        """Test an upload request with a filehandle."""
        # filehandle is a read()able
        fhdata = ['one', 'two']
        mock_fh = mock.Mock(read=mock.Mock(side_effect=fhdata))
        self._test_upload_request(mock_rq, mock_fh, fhdata)
        # Make sure the file handle's read method was invoked
        mock_fh.read.assert_has_calls([mock.call(65536)] * len(fhdata))

    def _assert_paths_equivalent(self, exp, act):
        """Ensures two paths or hrefs are "the same".

        Query parameter keys may be specified in any order, though their values
        must match exactly.  The rest of the path must be identical.

        :param exp: Expected path
        :param act: Actual path (produced by test)
        """
        p_exp = urlparse.urlparse(exp)
        p_act = urlparse.urlparse(act)
        self.assertEqual(p_exp.scheme, p_act.scheme)
        self.assertEqual(p_exp.netloc, p_act.netloc)
        self.assertEqual(p_exp.path, p_act.path)
        self.assertEqual(p_exp.fragment, p_act.fragment)
        qs_exp = urlparse.parse_qs(p_exp.query)
        qs_act = urlparse.parse_qs(p_act.query)
        for vals in qs_exp.values():
            vals.sort()
        for vals in qs_act.values():
            vals.sort()
        self.assertEqual(qs_exp, qs_act)

    @mock.patch('requests.Session')
    def test_extend_path(self, mock_session):
        # Init test data
        adapter = adp.Adapter(self.sess)

        path = adapter.extend_path('basepath', suffix_type='suffix',
                                   suffix_parm='suffix_parm',
                                   detail='detail',
                                   xag=[c.XAG.VIO_FMAP])

        expected_path = ('basepath/suffix/suffix_parm?detail=detail&'
                         'group=ViosFCMapping')
        self._assert_paths_equivalent(expected_path, path)

        # Multiple XAGs in a set
        path = adapter.extend_path('basepath', suffix_type='suffix',
                                   suffix_parm='suffix_parm',
                                   detail='detail',
                                   xag={c.XAG.VIO_FMAP, c.XAG.VIO_NET})

        expected_path = ('basepath/suffix/suffix_parm?detail=detail&'
                         'group=ViosFCMapping,ViosNetwork')
        self._assert_paths_equivalent(expected_path, path)

        # Verify sorting
        path = adapter.extend_path('basepath', suffix_type='suffix',
                                   suffix_parm='suffix_parm',
                                   detail='detail',
                                   xag=[c.XAG.VIO_NET, c.XAG.VIO_FMAP])

        expected_path = ('basepath/suffix/suffix_parm?detail=detail&'
                         'group=ViosFCMapping,ViosNetwork')
        self._assert_paths_equivalent(expected_path, path)

        # Explicitly no XAG
        path = adapter.extend_path('basepath', suffix_type='suffix',
                                   suffix_parm='suffix_parm', detail='detail',
                                   xag=[])

        expected_path = 'basepath/suffix/suffix_parm?detail=detail'
        self._assert_paths_equivalent(expected_path, path)

        # Ensure unspecified XAG defaults to group=None
        path = adapter.extend_path('basepath', suffix_type='suffix',
                                   suffix_parm='suffix_parm')

        expected_path = 'basepath/suffix/suffix_parm?group=None'
        self._assert_paths_equivalent(expected_path, path)

        # ...except for specific suffix types 'quick' and 'do'
        path = adapter.extend_path('basepath', suffix_type='quick',
                                   suffix_parm='suffix_parm')

        expected_path = 'basepath/quick/suffix_parm'
        self._assert_paths_equivalent(expected_path, path)

        path = adapter.extend_path('basepath', suffix_type='do',
                                   suffix_parm='suffix_parm')

        expected_path = 'basepath/do/suffix_parm'
        self._assert_paths_equivalent(expected_path, path)

        # Ensure arg xags and path xags interact correctly
        # path_xag=None, arg_xag=None => group=None
        self._assert_paths_equivalent(
            'basepath?group=None', adapter.extend_path('basepath'))
        # path_xag='None', arg_xag=None => group=None
        self._assert_paths_equivalent(
            'basepath?group=None', adapter.extend_path('basepath?group=None'))
        # path_xag='a,b,c', arg_xag=None => group=a,b,c
        self._assert_paths_equivalent(
            'basepath?group=a,b,c',
            adapter.extend_path('basepath?group=a,b,c'))
        # path_xag=None, arg_xag=() => no group=
        self._assert_paths_equivalent(
            'basepath', adapter.extend_path('basepath', xag=()))
        # path_xag='None', arg_xag={} => no group=
        self._assert_paths_equivalent(
            'basepath', adapter.extend_path('basepath?group=None', xag={}))
        # path_xag='a,b,c', arg_xag=[] => ValueError
        self.assertRaises(
            ValueError, adapter.extend_path, 'basepath?group=a,b,c', xag=[])
        # path_xag=None, arg_xag='a,b,c' => group='a,b,c'
        self._assert_paths_equivalent(
            'basepath?group=a,b,c',
            adapter.extend_path('basepath', xag={'a', 'b', 'c'}))
        # path_xag='None', arg_xag='a,b,c' => group='a,b,c'
        self._assert_paths_equivalent(
            'basepath?group=a,b,c',
            adapter.extend_path('basepath?group=None', xag=('a', 'b', 'c')))
        # path_xag='a,b,c', arg_xag='a,b,c' => group='a,b,c'
        self._assert_paths_equivalent(
            'basepath?group=a,b,c',
            adapter.extend_path('basepath?group=a,b,c', xag=['a', 'b', 'c']))
        # path_xag='a,b,c', arg_xag='d,e,f' => ValueError
        self.assertRaises(ValueError, adapter.extend_path,
                          'basepath?group=a,b,c', xag=['d', 'e', 'f'])
        # Multi-instance query params properly reassembled.
        self._assert_paths_equivalent(
            'basepath?foo=1,2,3&group=a,b,c&foo=4,5,6',
            adapter.extend_path('basepath?foo=4,5,6&group=None&foo=1,2,3',
                                xag=['a', 'b', 'c']))

        # Additional queryparams (add_qp)
        # Explicit None
        self._assert_paths_equivalent(
            'basepath', adapter.extend_path('basepath', xag=[], add_qp=None))
        # Proper escaping
        self._assert_paths_equivalent(
            'basepath?one=%23%24%25%5E%26',
            adapter.extend_path('basepath', xag=[], add_qp=[('one', '#$%^&')]))
        # Duplicated keys (order preserved) and proper handling of non-strings
        self._assert_paths_equivalent(
            'basepath?1=3&1=2',
            adapter.extend_path('basepath', xag=[], add_qp=[(1, 3), (1, 2)]))
        # Proper behavior combined with implicit xag
        self._assert_paths_equivalent(
            'basepath?group=None&key=value&something=else',
            adapter.extend_path(
                'basepath', add_qp=[('key', 'value'), ('something', 'else')]))
        # Combined with xags and an existing querystring
        self._assert_paths_equivalent(
            'basepath?already=here&group=a,b,c&key=value&something=else',
            adapter.extend_path(
                'basepath?already=here', xag=['a', 'b', 'c'],
                add_qp=[('key', 'value'), ('something', 'else')]))

    @mock.patch('pypowervm.adapter.LOG')
    @mock.patch('pypowervm.adapter.Adapter.read_by_path')
    def test_read_by_href(self, mock_read_by_path, mock_log):
        """Ensure read_by_href correctly extends, preserves query strings."""
        def validate_read_by_path(expected):
            def _read_by_path(path, etag, timeout, auditmemento, age,
                              sensitive, helpers):
                self._assert_paths_equivalent(expected, path)
                for param in (etag, auditmemento, helpers):
                    self.assertIsNone(param)
                for param2 in (age, timeout):
                    self.assertEqual(-1, param2)
                self.assertFalse(sensitive)
            return _read_by_path

        self.sess.host = 'foo'
        self.sess.port = 123
        adapter = adp.Adapter(self.sess)
        mock_read_by_path.side_effect = validate_read_by_path(
            '/rest/api/uom/Bar?k=v&group=None#frag')
        adapter.read_by_href('http://foo:123/rest/api/uom/Bar?k=v#frag')
        self.assertFalse(mock_log.debug.called)

        self.sess.host = 'bar'
        mock_read_by_path.side_effect = validate_read_by_path(
            '/rest/api/uom/Bar?k=v&group=None#frag')
        adapter.read_by_href('http://foo:123/rest/api/uom/Bar?k=v#frag')
        self.assertTrue(mock_log.debug.called)

        mock_read_by_path.side_effect = validate_read_by_path(
            '/rest/api/uom/Bar?k=v&group=RealGroup#frag')
        adapter.read_by_href(
            'http://foo:123/rest/api/uom/Bar?k=v&group=RealGroup#frag')

    @mock.patch('requests.Session')
    def test_delete(self, mock_session):
        """Test delete() method found in the Adapter class."""
        # Init test data
        root_type = 'ManagedSystem'
        root_id = 'id'
        adapter = adp.Adapter(self.sess)

        delete_response = self._mk_response(204)

        # Mock out the method and class we are not currently testing
        session = mock_session.return_value
        session.request.return_value = delete_response

        # Run the actual test
        ret_delete_value = adapter.delete(root_type, root_id)

        # Verify Correct path was built in build_path()
        reqpath = adp.Adapter.build_path('uom', root_type, root_id, xag=[])

        # Verify the return value
        # self.assertIsInstance(ret_delete_value, adp.Response)
        self.assertEqual('DELETE', ret_delete_value.reqmethod)
        self.assertEqual(204, ret_delete_value.status)
        self.assertEqual(reqpath, ret_delete_value.reqpath)

    @mock.patch.object(builtins, 'open')
    def test_auth_file_error(self, mock_open_patch):
        mock_open_patch.side_effect = IOError(errno.EACCES, 'Error')
        self.assertRaises(pvmex.AuthFileReadError,
                          self.sess._get_auth_tok_from_file,
                          mock.Mock(), mock.Mock())

        mock_open_patch.side_effect = IOError(errno.EIO, 'Error')
        self.assertRaises(pvmex.AuthFileAccessError,
                          self.sess._get_auth_tok_from_file,
                          mock.Mock(), mock.Mock())

    @mock.patch('pypowervm.adapter.LOG')
    @mock.patch('requests.Session')
    def test_unauthorized_error(self, mock_session, mock_log):
        """401 (unauthorized) calling Adapter.create()."""

        # Init test data
        adapter = adp.Adapter(self.sess)
        new_scsi = pvm_stor.VSCSIClientAdapterElement.bld(adapter)

        element = new_scsi
        root_type = 'ManagedSystem'
        root_id = 'id'
        child_type = 'LogicalPartition'

        create_response = self._mk_response(401)

        # Mock out the method and class we are not currently testing
        session = mock_session.return_value
        session.request.return_value = create_response

        # Run the actual test
        self.assertRaises(pvmex.HttpError, adapter.create, element,
                          root_type, root_id, child_type)
        self.assertEqual(1, mock_log.warning.call_count)

    def test_element_iter(self):
        """Test the ETElement iter() method found in the Adapter class."""

        # Init test data
        children = [ent.Element('Type1', None, text='T1_0'),
                    ent.Element('Type12', None, text='T12_0'),
                    ent.Element('Type1', None, text='T1_1'),
                    ent.Element('Type12', None, text='T12_1'),
                    ent.Element('Type1', None, text='T1_2')]
        top_element = ent.Element('Top', None,
                                  attrib={'schemaVersion': 'V1_0'},
                                  children=children)

        def _count_elem(top, tag, it=None, assert_tag=True):
            elem_count = 0
            it = it if it else top.iter(tag=tag)
            for elem in it:
                if assert_tag:
                    self.assertEqual(elem.tag, tag)
                elem_count += 1
            return elem_count

        # Run the actual tests

        # Ensure all elements are traversed if we don't specify a tag
        self.assertEqual(_count_elem(top_element, 'Type1',
                                     it=top_element.iter(),
                                     assert_tag=False), 6)

        # Ensure all elements are traversed for tag=*
        self.assertEqual(_count_elem(top_element, 'Type1',
                                     it=top_element.iter(tag='*'),
                                     assert_tag=False), 6)

        # Ensure all elements are traversed for tag=None
        self.assertEqual(_count_elem(top_element, 'Type1',
                                     it=top_element.iter(tag=None),
                                     assert_tag=False), 6)

        # Get only the Type1 elements
        self.assertEqual(_count_elem(top_element, 'Type1'), 3)

        # Get only the top
        self.assertEqual(_count_elem(top_element, 'Top'), 1)

    @mock.patch('pypowervm.entities.Feed.unmarshal_atom_feed')
    @mock.patch('pypowervm.entities.Entry.unmarshal_atom_entry')
    @mock.patch('lxml.etree.fromstring')
    def test_extract_atom(self, mock_fromstring, mock_unm_ent, mock_unm_feed):
        resp = adp.Response('meth', '/rest/api/uom/Debug/SetLoggingLevel',
                            'status', 'reason', 'headers', body='body')
        feed_ret = mock.Mock(tag=etree.QName(c.ATOM_NS, 'feed'))
        entry_ret = mock.Mock(tag=etree.QName(c.ATOM_NS, 'entry'))

        # Empty content; "Response is not an Atom feed/entry"
        mock_fromstring.return_value = None
        self.assertIsNotNone(resp._extract_atom())
        mock_fromstring.assert_called_with('body')
        mock_unm_feed.assert_not_called()
        mock_unm_ent.assert_not_called()

        # Unmarshal feed (returns None)
        mock_fromstring.return_value = feed_ret
        self.assertIsNone(resp._extract_atom())
        mock_unm_feed.assert_called_once_with(feed_ret, resp)
        mock_unm_ent.assert_not_called()
        mock_unm_feed.reset_mock()

        # Unmarshal entry (returns None)
        mock_fromstring.return_value = entry_ret
        self.assertIsNone(resp._extract_atom())
        mock_unm_ent.assert_called_once_with(entry_ret, resp)
        mock_unm_feed.assert_not_called()
        mock_unm_ent.reset_mock()

        # Unmarshal a 'Debug' response (returns None)
        mock_fromstring.return_value = mock.Mock(tag='debug output')
        self.assertIsNone(resp._extract_atom())
        mock_unm_feed.assert_not_called()
        mock_unm_ent.assert_not_called()

        # 'fromstring' raises.  Make sure the return message came from the
        # right place (will include the exception text)
        mock_fromstring.side_effect = Exception("test_extract_atom")
        self.assertIn("test_extract_atom", resp._extract_atom())
        mock_unm_feed.assert_not_called()
        mock_unm_ent.assert_not_called()


class TestElement(testtools.TestCase):
    def setUp(self):
        super(TestElement, self).setUp()
        self.adpt = self.useFixture(fx.AdapterFx()).adpt

    def test_cdata(self):
        no_cdata = ent.Element('tag', self.adpt, text='text', cdata=False)
        with_cdata = ent.Element('tag', self.adpt, text='text', cdata=True)
        self.assertEqual(
            no_cdata.toxmlstring(),
            '<uom:tag xmlns:uom="http://www.ibm.com/xmlns/systems/power/'
            'firmware/uom/mc/2012_10/">text</uom:tag>'.encode('utf-8'))
        self.assertEqual(
            with_cdata.toxmlstring(),
            '<uom:tag xmlns:uom="http://www.ibm.com/xmlns/systems/power/firmwa'
            're/uom/mc/2012_10/"><![CDATA[text]]></uom:tag>'.encode('utf-8'))

    def test_tag_namespace(self):
        el = ent.Element('tag', self.adpt)
        self.assertEqual(el.element.tag, '{http://www.ibm.com/xmlns/systems/po'
                                         'wer/firmware/uom/mc/2012_10/}tag')
        # entities.Element.tag strips the namespace
        self.assertEqual(el.tag, 'tag')
        self.assertEqual(el.namespace, 'http://www.ibm.com/xmlns/systems/powe'
                                       'r/firmware/uom/mc/2012_10/')
        # Test setter
        el.tag = 'gat'
        self.assertEqual(el.element.tag, '{http://www.ibm.com/xmlns/systems/po'
                                         'wer/firmware/uom/mc/2012_10/}gat')
        self.assertEqual(el.tag, 'gat')
        el.namespace = 'foo'
        self.assertEqual(el.namespace, 'foo')
        # Now with no namespace
        el = ent.Element('tag', self.adpt, ns='')
        self.assertEqual(el.element.tag, 'tag')
        self.assertEqual(el.tag, 'tag')
        self.assertEqual(el.namespace, '')
        el.tag = 'gat'
        self.assertEqual(el.element.tag, 'gat')
        self.assertEqual(el.tag, 'gat')
        el.namespace = 'foo'
        self.assertEqual(el.namespace, 'foo')


class TestAdapterClasses(subunit.IsolatedTestCase, testtools.TestCase):
    def setUp(self):
        super(TestAdapterClasses, self).setUp()
        self.mock_logoff = self.useFixture(
            fixtures.MockPatchObject(adp.Session, '_logoff')).mock
        self.mock_logon = self.useFixture(
            fixtures.MockPatchObject(adp.Session, '_logon')).mock
        self.mock_events = self.useFixture(
            fixtures.MockPatchObject(adp._EventListener, '_get_events')).mock
        # Mock the initial events coming in on start
        self.mock_events.return_value = {'general': 'init'}, [], []

    def test_instantiation(self):
        """Direct instantiation of EventListener is not allowed."""
        # Get a session
        sess = adp.Session()
        # Now get the EventListener
        self.assertRaises(TypeError, adp.EventListener, sess)

        # Mock the session token like we logged on
        sess._sessToken = 'token'.encode('utf-8')
        # Ensure we get an EventListener
        self.assertIsInstance(sess.get_event_listener(), adp.EventListener)

    def test_shutdown_session(self):
        """Test garbage collection of the session.

        Ensures the Session can be properly garbage collected.
        """
        # Get a session
        sess = adp.Session()
        # Mock the session token like we logged on
        sess._sessToken = 'token'.encode('utf-8')
        # It should have logged on but not off.
        self.assertTrue(self.mock_logon.called)
        self.assertFalse(self.mock_logoff.called)

        # Get an event listener to test the weak references
        event_listen = sess.get_event_listener()

        # Test the circular reference (but one link is weak)
        sess.hello = 'hello'
        self.assertEqual(sess.hello, event_listen.adp.session.hello)

        # There should be 1 reference to the session (ours)
        self.assertEqual(1, len(gc.get_referrers(sess)))

    def test_shutdown_adapter(self):
        """Test garbage collection of the session, event listener.

        Ensures the proper shutdown of the session and event listener when
        we start with constructing an Adapter, implicit session and
        EventListener.
        """
        # Get Adapter, implicit session
        adapter = adp.Adapter()
        adapter.session._sessToken = 'token'.encode('utf-8')
        # Get construct and event listener
        adapter.session.get_event_listener()

        # Turn off the event listener
        adapter.session.get_event_listener().shutdown()
        # Session is still active
        self.assertFalse(self.mock_logoff.called)

        # The only thing that refers the adapter is our reference
        self.assertEqual(1, len(gc.get_referrers(adapter)))


class TestElementInject(testtools.TestCase):

    def setUp(self):
        super(TestElementInject, self).setUp()
        self.adpt = self.useFixture(fx.AdapterFx()).adpt
        self.ordering_list = ('AdapterType', 'UseNextAvailableSlotID',
                              'RemoteLogicalPartitionID', 'RemoteSlotNumber')
        self.child_at = ent.Element('AdapterType', self.adpt, text='Client')
        self.child_unasi = ent.Element('UseNextAvailableSlotID', self.adpt,
                                       text='true')
        self.child_rlpi1 = ent.Element('RemoteLogicalPartitionID', self.adpt,
                                       text='1')
        self.child_rlpi2 = ent.Element('RemoteLogicalPartitionID', self.adpt,
                                       text='2')
        self.child_rlpi3 = ent.Element('RemoteLogicalPartitionID', self.adpt,
                                       text='3')
        self.child_rsn = ent.Element('RemoteSlotNumber', self.adpt,
                                     text='12')
        self.all_children = [
            self.child_at, self.child_unasi, self.child_rlpi1, self.child_rsn]

    def _mk_el(self, children):
        return ent.Element('VirtualSCSIClientAdapter', self.adpt,
                           attrib={'schemaVersion': 'V1_0'},
                           children=children)

    def assert_expected_children(self, parent, *expected_children):
        """Assert that *children are the children of parent, in that order.

        :param parent: Parent adapter.Element
        :param children: Child adapter.Elements
        """
        # etree.Element doesn't implement __eq__, so different instances of the
        # same Element aren't "equal".  Compare XML strings instead.
        actual = [etree.tostring(elem) for elem in list(parent.element)]
        expected = [etree.tostring(chld.element) for chld in expected_children]
        self.assertEqual(actual, expected)

    def test_no_children(self):
        """Inject when the element has no children - should "append"."""
        el = self._mk_el([])
        el.inject(self.child_rlpi1)
        self.assert_expected_children(el, self.child_rlpi1)
        # Result should be same regardless of other params
        el = self._mk_el([])
        el.inject(self.child_rlpi1, self.ordering_list, replace=False)
        self.assert_expected_children(el, self.child_rlpi1)

    def test_subelement_found_one_replace_true(self):
        """Replace existing child with same tag."""
        el = self._mk_el(self.all_children)
        el.inject(self.child_rlpi2, self.ordering_list)
        self.assert_expected_children(el, self.child_at, self.child_unasi,
                                      self.child_rlpi2, self.child_rsn)
        # Proving default replace=True - same result if specified
        el = self._mk_el(self.all_children)
        el.inject(self.child_rlpi2, self.ordering_list, replace=True)
        self.assert_expected_children(el, self.child_at, self.child_unasi,
                                      self.child_rlpi2, self.child_rsn)

    def test_subelement_found_mult_replace_true(self):
        """Replace existing child with same tag when >1 such children.

        Should replace the last such child.
        """
        el = self._mk_el([self.child_at, self.child_unasi, self.child_rlpi1,
                          self.child_rlpi3, self.child_rsn])
        el.inject(self.child_rlpi2, self.ordering_list)
        self.assert_expected_children(el, self.child_at, self.child_unasi,
                                      self.child_rlpi1, self.child_rlpi2,
                                      self.child_rsn)

    def test_subelement_found_replace_false(self):
        """Inject after existing child(ren) with same tag."""
        el = self._mk_el(self.all_children)
        el.inject(self.child_rlpi2, self.ordering_list, False)
        self.assert_expected_children(el, self.child_at, self.child_unasi,
                                      self.child_rlpi1, self.child_rlpi2,
                                      self.child_rsn)
        el.inject(self.child_rlpi3, self.ordering_list, False)
        self.assert_expected_children(el, self.child_at, self.child_unasi,
                                      self.child_rlpi1, self.child_rlpi2,
                                      self.child_rlpi3, self.child_rsn)

    def test_subelement_not_in_ordering_list(self):
        """Subelement not in ordering list - should append."""
        el = self._mk_el(self.all_children)
        ch = ent.Element('SomeNewElement', self.adpt, text='foo')
        el.inject(ch, ordering_list=self.ordering_list)
        self.assert_expected_children(el, self.child_at, self.child_unasi,
                                      self.child_rlpi1, self.child_rsn, ch)

    def test_first_populated(self):
        """Inject the first child when children are otherwise populated."""
        el = self._mk_el(self.all_children[1:])
        el.inject(self.child_at, self.ordering_list)
        self.assert_expected_children(el, self.child_at, self.child_unasi,
                                      self.child_rlpi1, self.child_rsn)

    def test_first_sparse(self):
        """Inject the first child when children are sparsely populated."""
        # This is most interesting when the existing child is not the one right
        # next to the injectee.
        el = self._mk_el([self.child_rlpi1])
        el.inject(self.child_at, self.ordering_list)
        self.assert_expected_children(el, self.child_at, self.child_rlpi1)

    def test_last_populated(self):
        """Inject the last child when children are otherwise populated."""
        el = self._mk_el(self.all_children[:-1])
        el.inject(self.child_rsn, self.ordering_list)
        self.assert_expected_children(el, self.child_at, self.child_unasi,
                                      self.child_rlpi1, self.child_rsn)

    def test_last_sparse(self):
        """Inject the last child when children are sparsely populated."""
        # This is most interesting when the existing child is not the one right
        # next to the injectee.
        el = self._mk_el([self.child_unasi])
        el.inject(self.child_rsn, self.ordering_list)
        self.assert_expected_children(el, self.child_unasi, self.child_rsn)

    def test_middle_populated(self):
        """Inject a middle child when children are otherwise populated."""
        el = self._mk_el([self.child_at, self.child_unasi, self.child_rsn])
        el.inject(self.child_rlpi1, self.ordering_list)
        self.assert_expected_children(el, self.child_at, self.child_unasi,
                                      self.child_rlpi1, self.child_rsn)

    def test_middle_sparse(self):
        """Inject a middle child when children are sparsely populated."""
        el = self._mk_el([self.child_at, self.child_rsn])
        el.inject(self.child_rlpi1, self.ordering_list)
        self.assert_expected_children(
            el, self.child_at, self.child_rlpi1, self.child_rsn)


class TestElementWrapper(testtools.TestCase):
    """Tests for the ElementWrapper class."""

    def setUp(self):
        super(TestElementWrapper, self).setUp()
        self.resp = pvmhttp.load_pvm_resp(NET_BRIDGE_FILE).get_response()
        self.nb1 = self.resp.feed.entries[0]
        self.resp2 = pvmhttp.load_pvm_resp(NET_BRIDGE_FILE).get_response()
        self.nb2 = self.resp2.feed.entries[0]

    def test_equality(self):
        """Validates that two elements loaded from the same data is equal."""
        sea1 = self._find_seas(self.nb1)[0]
        sea2 = self._find_seas(self.nb2)[0]
        self.assertTrue(sea1 == sea2)

        # Change the other SEA
        sea2.element.append(etree.Element('Bob'))
        self.assertFalse(sea1 == sea2)

    def test_inequality_by_subelem_change(self):
        sea1 = self._find_seas(self.nb1)[0]
        sea2 = self._find_seas(self.nb2)[0]
        sea_trunk = sea2.findall('TrunkAdapters/TrunkAdapter')[0]
        pvid = sea_trunk.find('PortVLANID')
        pvid.text = '1'
        self.assertFalse(sea1 == sea2)

    def _find_seas(self, entry):
        """Wrapper for the SEAs."""
        return entry.element.findall('SharedEthernetAdapters/'
                                     'SharedEthernetAdapter')



# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import unittest

import mock

import pypowervm.const as pc
import pypowervm.tests.test_fixtures as fx
import pypowervm.tests.test_utils.pvmhttp as pvmhttp
import pypowervm.tests.test_utils.test_wrapper_abc as twrap
import pypowervm.wrappers.base_partition as bp
import pypowervm.wrappers.logical_partition as lpar
import pypowervm.wrappers.network as net
import pypowervm.wrappers.virtual_io_server as vios

NET_BRIDGE_FILE = 'fake_network_bridge.txt'
VSWITCH_FEED_FILE = 'fake_vswitch_feed.txt'


class TestVNetwork(twrap.TestWrapper):

    file = 'fake_virtual_network_feed.txt'
    wrapper_class_to_test = net.VNet

    def test_vnet(self):
        self.assertEqual('https://9.1.2.3:12443/rest/api/uom/'
                         'ManagedSystem/67dca605-3923-34da-bd8f-26a378fc817f/'
                         'VirtualSwitch/ec8aaa54-9837-3c23-a541-a4e4be3ae489',
                         self.dwrap.associated_switch_uri)
        self.assertEqual('VLAN2227-ETHERNET0', self.dwrap.name)
        self.assertEqual(2227, self.dwrap.vlan)
        self.assertEqual(0, self.dwrap.vswitch_id)
        self.assertEqual(False, self.dwrap.tagged)

    def test_name(self):
        self.dwrap.name = 'Test'
        self.assertEqual('Test', self.dwrap.name)

    def test_vnet_new(self):
        """Tests the method that returns a VNet ElementWrapper."""
        vn_w = net.VNet.bld(None, 'name', 10, 'vswitch_uri', True)
        self.assertEqual('name', vn_w.name)
        self.assertEqual(10, vn_w.vlan)
        self.assertTrue(vn_w.tagged)

    def test_wrapper_class(self):
        self.assertEqual(net.VNet.schema_type, 'VirtualNetwork')
        self.assertEqual(net.VNet.schema_ns, pc.UOM_NS)
        self.assertTrue(net.VNet.has_metadata)
        self.assertEqual(net.VNet.default_attrib, pc.DEFAULT_SCHEMA_ATTR)


class TestVSwitch(twrap.TestWrapper):

    file = 'fake_vswitch_feed.txt'
    wrapper_class_to_test = net.VSwitch

    def test_bld(self):
        """Tests that the vSwitch element can be built."""
        vs = net.VSwitch.bld(None, 'Test')
        self.assertEqual('Test', vs.name)
        self.assertEqual(net.VSwitchMode.VEB, vs.mode)
        self.assertListEqual([], vs.vnet_uri_list)

        vs = net.VSwitch.bld(None, 'Test', net.VSwitchMode.VEPA)
        self.assertEqual('Test', vs.name)
        self.assertEqual(net.VSwitchMode.VEPA, vs.mode)
        self.assertListEqual([], vs.vnet_uri_list)

    def test_feed(self):
        """Tests the feed of virtual switches."""
        vswitches = net.VSwitch.wrap(self.resp)
        self.assertTrue(len(vswitches) >= 1)
        for vswitch in vswitches:
            self.assertIsNotNone(vswitch.etag)

    def test_data(self):
        self.assertEqual('ETHERNET0', self.dwrap.name)
        self.assertEqual(0, self.dwrap.switch_id)
        self.assertEqual('Veb', self.dwrap.mode)
        self.assertEqual('https://9.1.2.3:12443/rest/api/uom/ManagedSystem/'
                         '4abca7ff-3710-3160-b9e4-cb4456c33f43/VirtualSwitch/'
                         '4d9735ae-feaf-32c2-a1bc-102026df9168?group=None',
                         self.dwrap.href)
        self.assertEqual('https://9.1.2.3:12443/rest/api/uom/ManagedSystem/'
                         '4abca7ff-3710-3160-b9e4-cb4456c33f43/VirtualSwitch/'
                         '4d9735ae-feaf-32c2-a1bc-102026df9168',
                         self.dwrap.related_href)

    def test_wrapper_class(self):
        self.assertEqual(net.VSwitch.schema_type, 'VirtualSwitch')
        self.assertEqual(net.VSwitch.schema_ns, pc.UOM_NS)
        self.assertTrue(net.VSwitch.has_metadata)
        self.assertEqual(net.VSwitch.default_attrib, pc.DEFAULT_SCHEMA_ATTR)

    def test_set_mode(self):
        """Tests that the vSwitch element can have the mode set."""
        vs = net.VSwitch.bld(None, 'Test')
        self.assertEqual(net.VSwitchMode.VEB, vs.mode)
        vs.mode = net.VSwitchMode.VEPA
        self.assertEqual(net.VSwitchMode.VEPA, vs.mode)
        vs.mode = net.VSwitchMode.VEB
        self.assertEqual(net.VSwitchMode.VEB, vs.mode)


class TestLoadGroup(unittest.TestCase):
    def test_wrapper_class(self):
        self.assertEqual(net.LoadGroup.schema_type, 'LoadGroup')
        self.assertEqual(net.LoadGroup.schema_ns, pc.UOM_NS)
        self.assertTrue(net.LoadGroup.has_metadata)
        self.assertEqual(net.LoadGroup.default_attrib, pc.DEFAULT_SCHEMA_ATTR)


class TestTrunkAdapter(unittest.TestCase):
    def test_wrapper_class(self):
        self.assertEqual(net.TrunkAdapter.schema_type, 'TrunkAdapter')
        self.assertEqual(net.TrunkAdapter.schema_ns, pc.UOM_NS)
        self.assertFalse(net.TrunkAdapter.has_metadata)
        self.assertEqual(net.TrunkAdapter.default_attrib,
                         pc.DEFAULT_SCHEMA_ATTR)


class TestSEA(unittest.TestCase):
    def test_wrapper_class(self):
        self.assertEqual(net.SEA.schema_type, 'SharedEthernetAdapter')
        self.assertEqual(net.SEA.schema_ns, pc.UOM_NS)
        self.assertTrue(net.SEA.has_metadata)
        self.assertEqual(net.SEA.default_attrib, pc.DEFAULT_SCHEMA_ATTR)


class TestNetBridge(unittest.TestCase):
    def test_wrapper_class(self):
        self.assertEqual(net.NetBridge.schema_type, 'NetworkBridge')
        self.assertEqual(net.NetBridge.schema_ns, pc.UOM_NS)
        self.assertTrue(net.NetBridge.has_metadata)
        self.assertEqual(net.NetBridge.default_attrib, pc.DEFAULT_SCHEMA_ATTR)


class TestNetwork(twrap.TestWrapper):

    file = 'fake_network_bridge.txt'
    wrapper_class_to_test = net.NetBridge
    mock_adapter_fx_args = dict(traits=fx.LocalPVMTraits)

    def set_vnet(self, aware):
        # Since they're all references through the same adapter, setting traits
        # on dwrap's element's adapter ought to affect all sub-elements, etc.
        self.adptfx.set_traits(fx.RemoteHMCTraits if aware
                               else fx.RemotePVMTraits)

    def test_pvid(self):
        self.assertEqual(1, self.dwrap.pvid)

    def test_configuration_state(self):
        self.assertEqual(net.SEAState.CONFIGURED,
                         self.dwrap.seas[0].configuration_state)

    def test_load_balance(self):
        self.assertTrue(self.dwrap.load_balance)

    def test_uuid(self):
        self.assertEqual(
            '764f3423-04c5-3b96-95a3-4764065400bd', self.dwrap.uuid)

    def test_vnet_uri_list(self):
        uri_list = self.dwrap.vnet_uri_list
        self.assertEqual(13, len(uri_list))
        self.assertEqual('http', uri_list[0][:4])

    def test_contrl_channel(self):
        vios_file = pvmhttp.PVMFile('fake_vios_feed.txt')

        vios_resp = pvmhttp.PVMResp(pvmfile=vios_file).get_response()
        vios_wrap = vios.VIOS.wrap(vios_resp.feed.entries[0])

        self.assertEqual('ent5', vios_wrap.seas[0].control_channel)

    def test_contrl_channel_id(self):
        self.assertEqual(99, self.dwrap.control_channel_id)

    def test_crt_net_bridge(self):
        vswitch_file = pvmhttp.PVMFile('fake_vswitch_feed.txt')

        vswitch_resp = pvmhttp.PVMResp(pvmfile=vswitch_file).get_response()
        vsw_wrap = net.VSwitch.wrap(vswitch_resp.feed.entries[0])

        # Create mocked data
        nb = net.NetBridge.bld(self.adpt, pvid=1,
                               vios_to_backing_adpts=[('vio_href1', 'ent0'),
                                                      ('vio_href2', 'ent2')],
                               vswitch=vsw_wrap, load_balance=True)

        self.assertIsNotNone(nb)
        self.assertEqual(1, nb.pvid)
        self.assertEqual(2, len(nb.seas))
        self.assertEqual(0, len(nb.load_grps))
        self.assertTrue(nb.load_balance)

        # First SEA.  Should be the primary
        sea1 = nb.seas[0]
        self.assertIsNotNone(sea1)
        self.assertEqual(1, sea1.pvid)
        self.assertEqual('vio_href1', sea1.vio_uri)
        self.assertEqual('ent0', sea1.backing_device.dev_name)
        self.assertTrue(sea1.is_primary)

        # Validate the trunk.
        ta = sea1.primary_adpt
        self.assertTrue(ta._required)
        self.assertEqual(1, ta.pvid)
        self.assertFalse(ta.has_tag_support)
        self.assertEqual(vsw_wrap.switch_id, ta.vswitch_id)
        self.assertEqual(1, ta.trunk_pri)
        self.assertEqual(vsw_wrap.related_href, ta.associated_vswitch_uri)

        # Check that the second SEA is similar but not primary.
        sea2 = nb.seas[1]
        self.assertIsNotNone(sea2)
        self.assertEqual(1, sea2.pvid)
        self.assertEqual('vio_href2', sea2.vio_uri)
        self.assertEqual('ent2', sea2.backing_device.dev_name)
        self.assertFalse(sea2.is_primary)
        self.assertIsNone(sea2.ha_mode)

        # Validate the second SEA trunk.
        ta = sea2.primary_adpt
        self.assertTrue(ta._required)
        self.assertEqual(1, ta.pvid)
        self.assertFalse(ta.has_tag_support)
        self.assertEqual(vsw_wrap.switch_id, ta.vswitch_id)
        self.assertEqual(2, ta.trunk_pri)
        self.assertEqual(vsw_wrap.related_href, ta.associated_vswitch_uri)

    def test_crt_sea(self):
        vswitch_file = pvmhttp.PVMFile('fake_vswitch_feed.txt')

        vswitch_resp = pvmhttp.PVMResp(pvmfile=vswitch_file).get_response()
        vsw_wrap = net.VSwitch.wrap(vswitch_resp.feed.entries[0])

        # Create mocked data
        sea = net.SEA.bld(self.adpt, pvid=1, vios_href='127.0.0.1',
                          adpt_name='ent0', vswitch=vsw_wrap)

        self.assertIsNotNone(sea)
        self.assertEqual(1, sea.pvid)
        self.assertEqual('127.0.0.1', sea.vio_uri)
        self.assertEqual('ent0',
                         sea.backing_device.dev_name)

        ta = sea.primary_adpt
        self.assertTrue(ta._required)
        self.assertEqual(1, ta.pvid)
        self.assertFalse(ta.has_tag_support)
        self.assertEqual(vsw_wrap.switch_id, ta.vswitch_id)
        self.assertEqual(1, ta.trunk_pri)
        self.assertEqual(vsw_wrap.related_href, ta.associated_vswitch_uri)

    def test_crt_trunk_adapter(self):
        vswitch_file = pvmhttp.PVMFile('fake_vswitch_feed.txt')

        vswitch_resp = pvmhttp.PVMResp(pvmfile=vswitch_file).get_response()
        vsw_wrap = net.VSwitch.wrap(vswitch_resp.feed.entries[0])

        # Create mocked data
        ta = net.TrunkAdapter.bld(self.adpt, pvid=1, vlan_ids=[1, 2, 3],
                                  vswitch=vsw_wrap)

        self.assertIsNotNone(ta)
        self.assertTrue(ta._required)
        self.assertEqual(1, ta.pvid)
        self.assertEqual([1, 2, 3], ta.tagged_vlans)
        self.assertTrue(ta.has_tag_support)
        self.assertEqual(vsw_wrap.switch_id, ta.vswitch_id)
        self.assertEqual(1, ta.trunk_pri)
        self.assertEqual(vsw_wrap.related_href, ta.associated_vswitch_uri)

        # Try adding a VLAN to the trunk adapter.
        ta.tagged_vlans.append(4)
        self.assertEqual([1, 2, 3, 4], ta.tagged_vlans)

    def test_crt_load_group(self):
        # Create my mocked data
        uri_list = ['a', 'b', 'c']
        pvid = 1
        lg = net.LoadGroup.bld(self.adpt, pvid, uri_list)

        # Validate the data back
        self.assertIsNotNone(lg)
        self.assertEqual(1, lg.pvid)
        self.assertEqual(3, len(lg.vnet_uri_list))
        self.assertEqual('a', lg.vnet_uri_list[0])
        self.assertEqual('b', lg.vnet_uri_list[1])
        self.assertEqual('c', lg.vnet_uri_list[2])

    def test_load_groups(self):
        prim_ld_grp = self.dwrap.load_grps[0]
        self.assertIsNotNone(prim_ld_grp)
        self.assertEqual(1, prim_ld_grp.pvid)
        self.assertEqual(1, len(prim_ld_grp.trunk_adapters))
        self.assertEqual('U8246.L2C.0604C7A-V4-C2',
                         prim_ld_grp.trunk_adapters[0].loc_code)
        self.assertEqual(4, prim_ld_grp.trunk_adapters[0].vios_id)

        addl_ld_grps = self.dwrap.load_grps[1:]
        self.assertIsNotNone(addl_ld_grps)
        self.assertEqual(1, len(addl_ld_grps))

        self.assertEqual(
            12, len(addl_ld_grps[0].vnet_uri_list))
        addl_ld_grps[0].vnet_uri_list.append('fake_uri')
        self.assertEqual(
            13, len(addl_ld_grps[0].vnet_uri_list))
        addl_ld_grps[0].vnet_uri_list.remove('fake_uri')
        self.assertEqual(
            12, len(addl_ld_grps[0].vnet_uri_list))

        # Make sure that the reference to the Network Bridge is there.
        self.assertEqual(self.dwrap, prim_ld_grp._nb_root)

    def test_load_group_modification(self):
        """Verifies that the callbacks to the Network Bridge work.

        When modifying the Virtual Network list in the Load Group, those
        updates should be reflected back into the Network Bridge.
        """
        orig_len = len(self.dwrap.vnet_uri_list)
        ld_grp = self.dwrap.load_grps[0]
        lg_vnets = ld_grp.vnet_uri_list
        first_vnet = lg_vnets[0]
        lg_vnets.remove(first_vnet)

        self.assertEqual(orig_len - 1,
                         len(self.dwrap.vnet_uri_list))

    def test_sea_modification(self):
        """Verifies that the SEA can have a Trunk Adapter added to it."""
        vswitch_file = pvmhttp.PVMFile('fake_vswitch_feed.txt')

        vswitch_resp = pvmhttp.PVMResp(pvmfile=vswitch_file).get_response()
        vsw_wrap = net.VSwitch.wrap(vswitch_resp.feed.entries[0])

        # Create mocked data
        ta = net.TrunkAdapter.bld(
            self.adpt, pvid=1, vlan_ids=[1, 2, 3], vswitch=vsw_wrap)
        self.assertEqual(1, len(self.dwrap.seas[0].addl_adpts))
        self.dwrap.seas[0].addl_adpts.append(ta)
        self.assertEqual(2, len(self.dwrap.seas[0].addl_adpts))

        # Check that the total trunks is now three elements
        self.assertEqual(3, len(self.dwrap.seas[0]._get_trunks()))

    def test_supports_vlan(self):
        """Tests the supports_vlan method."""
        # Both styles should produce similar results.
        vnet_paths = [False, True]
        for use_vnet in vnet_paths:
            self.set_vnet(use_vnet)
            # PVID of primary adapter
            self.assertTrue(self.dwrap.supports_vlan(1))
            self.assertTrue(self.dwrap.supports_vlan("1"))

            # PVID of second adapter
            self.assertFalse(self.dwrap.supports_vlan(4094))
            self.assertFalse(self.dwrap.supports_vlan("4094"))

            # Additional VLAN of second adapter.
            self.assertTrue(self.dwrap.supports_vlan(100))
            self.assertTrue(self.dwrap.supports_vlan("100"))
            self.assertTrue(self.dwrap.supports_vlan(2228))
            self.assertTrue(self.dwrap.supports_vlan("2228"))
            self.assertTrue(self.dwrap.supports_vlan(2227))
            self.assertTrue(self.dwrap.supports_vlan("2227"))

            # A VLAN that isn't anywhere
            self.assertFalse(self.dwrap.supports_vlan(123))

    def test_supports_vlan_no_vnet(self):
        """Tests that a VLAN change affects trunks, not vnets."""
        self.dwrap.seas[0].primary_adpt.tagged_vlans.append(128)
        self.set_vnet(False)
        self.assertTrue(self.dwrap.supports_vlan(128))
        self.set_vnet(True)
        self.assertFalse(self.dwrap.supports_vlan(128))

    def test_no_primary_adpt(self):
        """Tests rare case that SEA has no primary adapter."""
        # Test to make sure None reference error is not hit
        self.assertIsNone(self.dwrap.seas[1].primary_adpt)
        self.assertEqual(self.dwrap.seas[1].addl_adpts, [])
        self.assertFalse(self.dwrap.seas[1].contains_device('abcd'))
        ct_ch = self.dwrap.seas[1].control_channel
        self.assertTrue(self.dwrap.seas[1].contains_device(ct_ch))

    def test_vswitch_id(self):
        """Tests that the pass thru of the vswitch id works."""
        self.assertEqual(2, self.dwrap.vswitch_id)

    def test_arbitrary_pvids(self):
        self.set_vnet(False)
        self.assertEqual([4094], self.dwrap.arbitrary_pvids)
        self.set_vnet(True)
        self.assertEqual([4094], self.dwrap.arbitrary_pvids)

    def test_list_vlans(self):
        # Both styles should produce similar results.
        vnet_paths = [False, True]
        for use_vnet in vnet_paths:
            self.set_vnet(use_vnet)

            # 1 is the PVID.  4094 is the arbitrary (only one arbitrary)
            val = set(self.dwrap.list_vlans())
            self.assertEqual({100, 150, 175, 200, 250, 300, 333, 350, 900,
                              1001, 2227, 2228, 1}, val)

            val = set(self.dwrap.list_vlans(pvid=False, arbitrary=True))
            self.assertEqual({4094, 100, 150, 175, 200, 250, 300, 333, 350,
                              900, 1001, 2227, 2228}, val)

            val = set(self.dwrap.list_vlans(pvid=False))
            self.assertEqual({100, 150, 175, 200, 250, 300, 333, 350, 900,
                              1001, 2227, 2228}, val)

            val = set(self.dwrap.list_vlans(arbitrary=True))
            self.assertEqual({1, 4094, 100, 150, 175, 200, 250, 300, 333, 350,
                              900, 1001, 2227, 2228}, val)

    def test_list_vlan_no_vnet(self):
        """Tests that a VLAN change affects trunks, not vnets."""
        self.dwrap.seas[0].primary_adpt.tagged_vlans.append(128)
        self.set_vnet(False)
        self.assertIn(128, self.dwrap.list_vlans())
        self.set_vnet(True)
        self.assertNotIn(128, self.dwrap.list_vlans())

    def test_seas(self):
        self.assertEqual(2, len(self.dwrap.seas))

        sea = self.dwrap.seas[0]

        # Test some properties
        self.assertEqual(1, sea.pvid)
        self.assertEqual('https://9.1.2.3:12443/rest/api/uom/ManagedSystem/'
                         '726e9cb3-6576-3df5-ab60-40893d51d074/VirtualIOServer'
                         '/691019AF-506A-4896-AADE-607E21FA93EE',
                         sea.vio_uri)
        self.assertEqual('ent8', sea.dev_name)
        self.assertEqual(net.HAMode.DISABLED, sea.ha_mode)

        new_sea = copy.deepcopy(sea)
        self.dwrap.seas.append(new_sea)

        self.assertEqual(3, len(self.dwrap.seas))

        sea_copy = copy.copy(self.dwrap.seas)
        sea_copy.remove(new_sea)
        self.dwrap.seas = sea_copy
        self.assertEqual(2, len(self.dwrap.seas))

        # Test the 'contains_device' method within the SEA.
        self.assertTrue(new_sea.contains_device('ent5'))
        self.assertFalse(new_sea.contains_device('ent2'))

    def test_sea_trunks(self):
        """Tests the trunk adapters on the SEA."""
        sea = self.dwrap.seas[0]

        # The primary adapter testing
        prim_t = sea.primary_adpt
        self.assertIsNotNone(prim_t)
        self.assertEqual(1, prim_t.pvid)
        self.assertFalse(prim_t.has_tag_support)
        self.assertEqual(0, len(prim_t.tagged_vlans))
        self.assertEqual(2, prim_t.vswitch_id)
        self.assertEqual('ent4', prim_t.dev_name)
        self.assertEqual(1, prim_t.trunk_pri)

        # The secondary adapter.
        addl_adpts = sea.addl_adpts
        self.assertIsNotNone(addl_adpts)
        self.assertEqual(1, len(addl_adpts))
        addl_adpt = addl_adpts[0]
        self.assertEqual(4094, addl_adpt.pvid)
        self.assertTrue(addl_adpt.has_tag_support)
        self.assertEqual(12, len(addl_adpt.tagged_vlans))
        self.assertEqual(2, addl_adpt.vswitch_id)
        self.assertEqual('ent5', addl_adpt.dev_name)
        self.assertEqual(1, addl_adpt.trunk_pri)

        # Try setting the tagged vlans
        orig_vlans = copy.copy(addl_adpt.tagged_vlans)
        addl_adpt.tagged_vlans.append(5)
        self.assertEqual(13, len(addl_adpt.tagged_vlans))
        addl_adpt.tagged_vlans = [1]
        self.assertEqual(1, len(addl_adpt.tagged_vlans))
        addl_adpt.tagged_vlans = orig_vlans
        self.assertEqual(12, len(addl_adpt.tagged_vlans))

        # Modify the tag support
        addl_adpt.has_tag_support = False
        self.assertFalse(addl_adpt.has_tag_support)
        addl_adpt.has_tag_support = True
        self.assertTrue(addl_adpt.has_tag_support)

    def test_varied_on(self):
        self.assertEqual(2, len(self.dwrap.seas))

        sea = self.dwrap.seas[0]

        # Try the varied_on property
        prim_t = sea.primary_adpt
        self.assertTrue(prim_t.varied_on)


class TestCNAWrapper(twrap.TestWrapper):

    file = 'fake_cna.txt'
    wrapper_class_to_test = net.CNA
    mock_adapter_fx_args = dict(traits=fx.LocalPVMTraits)

    def setUp(self):
        super(TestCNAWrapper, self).setUp()
        self.assertIsNotNone(self.entries.etag)

    def test_standard_crt(self):
        """Tests a standard create of the CNA."""
        test = net.CNA.bld(self.adpt, 1, "fake_vs")
        self.assertEqual('fake_vs', test.vswitch_uri)
        self.assertFalse(test.is_tagged_vlan_supported)
        self.assertEqual([], test.tagged_vlans)
        self.assertIsNotNone(test._use_next_avail_slot_id)
        self.assertTrue(test._use_next_avail_slot_id)
        self.assertIsNone(test.mac)
        self.assertIsNone(test.vsi_type_id)
        self.assertIsNone(test.vsi_type_version)
        self.assertIsNone(test.vsi_type_manager_id)
        self.assertIsNone(test.vswitch_id)
        self.assertEqual(1, test.pvid)
        self.assertNotIn(net._TA_TRUNK_PRI, str(test.toxmlstring()))
        self.assertFalse(test.is_trunk)
        self.assertIsNone(test.trunk_pri)
        self.assertFalse(test.enabled)

    def test_trunk_crt(self):
        """Tests a standard create of the CNA."""
        test = net.CNA.bld(self.adpt, 1, "fake_vs", trunk_pri=2)
        self.assertEqual('fake_vs', test.vswitch_uri)
        self.assertFalse(test.is_tagged_vlan_supported)
        self.assertEqual([], test.tagged_vlans)
        self.assertIsNotNone(test._use_next_avail_slot_id)
        self.assertTrue(test._use_next_avail_slot_id)
        self.assertIsNone(test.mac)
        self.assertIsNone(test.vsi_type_id)
        self.assertIsNone(test.vsi_type_version)
        self.assertIsNone(test.vsi_type_manager_id)
        self.assertIsNone(test.vswitch_id)
        self.assertEqual(1, test.pvid)
        self.assertIn(net._TA_TRUNK_PRI, str(test.toxmlstring()))
        self.assertTrue(test.is_trunk)
        self.assertEqual(test.trunk_pri, 2)

    def test_unique_crt(self):
        """Tests the create path with a non-standard flow for the CNA."""
        test = net.CNA.bld(
            self.adpt, 5, "fake_vs", mac_addr="aa:bb:cc:dd:ee:ff", slot_num=5,
            addl_tagged_vlans=[6, 7, 8, 9])
        self.assertEqual('fake_vs', test.vswitch_uri)
        self.assertTrue(test.is_tagged_vlan_supported)
        self.assertEqual([6, 7, 8, 9], test.tagged_vlans)
        self.assertEqual(5, test.slot)
        self.assertFalse(test._use_next_avail_slot_id)
        self.assertIsNotNone(test.mac)
        self.assertEqual("AABBCCDDEEFF", test.mac)
        self.assertEqual(5, test.pvid)
        self.assertIsNone(test.vsi_type_id)
        self.assertIsNone(test.vsi_type_version)
        self.assertIsNone(test.vsi_type_manager_id)
        self.assertIsNone(test.vswitch_id)
        self.assertNotIn(net._TA_TRUNK_PRI, str(test.toxmlstring()))
        self.assertFalse(test.is_trunk)
        self.assertIsNone(test.trunk_pri)

    def test_unasi_field(self):
        """UseNextAvailable(High)SlotID field is (not) used, as appropriate."""
        mock_vswitch = mock.Mock()
        mock_vswitch.related_href = 'href'
        # Do TrunkAdapter as well as CNA here
        # Traits fixture starts off "PVM" - should use High
        cna = net.CNA.bld(self.adpt, 1, "fake_vs")
        self.assertIsNone(cna._find(net._USE_NEXT_AVAIL_SLOT))
        self.assertIsNotNone(cna._find(net._USE_NEXT_AVAIL_HIGH_SLOT))
        self.assertTrue(cna._use_next_avail_slot_id)
        ta = net.TrunkAdapter.bld(self.adpt, 1, [], mock_vswitch)
        self.assertIsNone(ta._find(net._USE_NEXT_AVAIL_SLOT))
        self.assertIsNotNone(ta._find(net._USE_NEXT_AVAIL_HIGH_SLOT))
        self.assertEqual('Unknown', ta.dev_name)

        # When slot specified, no UseNextAvailable(High)SlotID
        cna = net.CNA.bld(self.adpt, 1, "fake_vs", slot_num=1)
        self.assertIsNone(cna._find(net._USE_NEXT_AVAIL_SLOT))
        self.assertIsNone(cna._find(net._USE_NEXT_AVAIL_HIGH_SLOT))
        self.assertFalse(cna._use_next_avail_slot_id)

        # Swap to HMC - should *not* use High
        self.adptfx.set_traits(fx.RemoteHMCTraits)
        cna = net.CNA.bld(self.adpt, 1, "fake_vs")
        self.assertIsNone(cna._find(net._USE_NEXT_AVAIL_HIGH_SLOT))
        self.assertIsNotNone(cna._find(net._USE_NEXT_AVAIL_SLOT))
        self.assertTrue(cna._use_next_avail_slot_id)
        ta = net.TrunkAdapter.bld(self.adpt, 1, [], mock_vswitch)
        self.assertIsNone(ta._find(net._USE_NEXT_AVAIL_HIGH_SLOT))
        self.assertIsNotNone(ta._find(net._USE_NEXT_AVAIL_SLOT))
        self.assertEqual('Unknown', cna.dev_name)

        # When slot specified, no UseNextAvailable(High)SlotID
        cna = net.CNA.bld(self.adpt, 1, "fake_vs", slot_num=1)
        self.assertIsNone(cna._find(net._USE_NEXT_AVAIL_SLOT))
        self.assertIsNone(cna._find(net._USE_NEXT_AVAIL_HIGH_SLOT))
        self.assertFalse(cna._use_next_avail_slot_id)

    @mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.create')
    @mock.patch('pypowervm.wrappers.logical_partition.LPAR.get')
    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.get')
    def test_cna_create(self, mock_vget, mock_lget, mock_ewrap_create):
        """CNA.create hack that mucks with UseNextAvailable(High)SlotID."""
        lpar_parent = mock.Mock(env=bp.LPARType.AIXLINUX,
                                is_mgmt_partition=False)
        vios_parent = mock.Mock(env=bp.LPARType.VIOS, is_mgmt_partition=False)
        mgmt_parent = mock.Mock(env=bp.LPARType.AIXLINUX,
                                is_mgmt_partition=True)
        sde_parent = mock.Mock(env=bp.LPARType.VIOS, is_mgmt_partition=True)

        # Exception paths for invalid parent spec
        cna = net.CNA.bld(self.adpt, 1, 'href')
        self.assertRaises(ValueError, cna.create)
        self.assertRaises(ValueError, cna.create, parent_type='foo')
        self.assertRaises(ValueError, cna.create, parent_uuid='foo')
        mock_ewrap_create.assert_not_called()
        mock_vget.assert_not_called()
        mock_lget.assert_not_called()

        # No parent, string parent_type gets converted.  Validate element
        # twiddling for VIOS and mgmt
        mock_lget.return_value = mgmt_parent
        mock_vget.return_value = vios_parent
        for ptyp, mck in ((lpar.LPAR, mock_lget), (vios.VIOS, mock_vget)):
            cna = net.CNA.bld(self.adpt, 1, 'href')
            self.assertEqual(
                mock_ewrap_create.return_value,
                cna.create(parent_type=ptyp.schema_type, parent_uuid='puuid'))
            mock_ewrap_create.assert_called_once_with(
                parent_type=ptyp, parent_uuid='puuid', timeout=-1,
                parent=mck.return_value)
            # One mck should get called in each loop
            mck.assert_called_once_with(self.adpt, uuid='puuid')
            # Element should get twiddled each time
            self.assertIsNone(cna._find(net._USE_NEXT_AVAIL_HIGH_SLOT))
            self.assertIsNotNone(cna._find(net._USE_NEXT_AVAIL_SLOT))
            self.assertTrue(cna._use_next_avail_slot_id)

            mock_ewrap_create.reset_mock()

        mock_lget.reset_mock()
        mock_vget.reset_mock()

        # No parent, wrapper parent_type, element twiddling for SDE (VIOS+mgmt)
        mock_vget.return_value = sde_parent
        cna = net.CNA.bld(self.adpt, 1, 'href')
        self.assertEqual(
            mock_ewrap_create.return_value,
            cna.create(parent_type=vios.VIOS, parent_uuid='puuid'))
        mock_ewrap_create.assert_called_once_with(
            parent_type=vios.VIOS, parent_uuid='puuid', timeout=-1,
            parent=mock_vget.return_value)
        mock_vget.assert_called_once_with(self.adpt, uuid='puuid')
        mock_lget.assert_not_called()
        # Element should get twiddled.
        self.assertIsNone(cna._find(net._USE_NEXT_AVAIL_HIGH_SLOT))
        self.assertIsNotNone(cna._find(net._USE_NEXT_AVAIL_SLOT))
        self.assertTrue(cna._use_next_avail_slot_id)

        mock_ewrap_create.reset_mock()
        mock_vget.reset_mock()

        # Parent specified, no element twiddling for plain LPAR
        cna = net.CNA.bld(self.adpt, 1, 'href')
        self.assertEqual(mock_ewrap_create.return_value,
                         cna.create(parent=lpar_parent))
        mock_ewrap_create.assert_called_once_with(
            parent_type=None, parent_uuid=None, timeout=-1, parent=lpar_parent)
        mock_vget.assert_not_called()
        mock_lget.assert_not_called()
        # Element should not get twiddled.
        self.assertIsNotNone(cna._find(net._USE_NEXT_AVAIL_HIGH_SLOT))
        self.assertIsNone(cna._find(net._USE_NEXT_AVAIL_SLOT))
        self.assertTrue(cna._use_next_avail_slot_id)

        mock_ewrap_create.reset_mock()

        # If slot specified, we skip the whole hack
        self.adptfx.set_traits(fx.RemoteHMCTraits)
        cna = net.CNA.bld(self.adpt, 1, 'href', slot_num=1)
        self.assertEqual(mock_ewrap_create.return_value,
                         cna.create(parent_type='ptyp', parent_uuid='puuid'))
        mock_ewrap_create.assert_called_once_with(
            parent_type='ptyp', parent_uuid='puuid', timeout=-1, parent=None)
        mock_vget.assert_not_called()
        mock_lget.assert_not_called()

        mock_ewrap_create.reset_mock()

        # For HMC, we skip the whole hack
        self.adptfx.set_traits(fx.RemoteHMCTraits)
        cna = net.CNA.bld(self.adpt, 1, 'href')
        self.assertEqual(mock_ewrap_create.return_value,
                         cna.create(parent_type='ptyp', parent_uuid='puuid'))
        mock_ewrap_create.assert_called_once_with(
            parent_type='ptyp', parent_uuid='puuid', timeout=-1, parent=None)
        mock_vget.assert_not_called()
        mock_lget.assert_not_called()

    def test_attrs(self):
        """Test getting the attributes."""
        self.assertEqual(32, self.dwrap.slot)
        self.assertEqual("FAD4433ED120", self.dwrap.mac)
        self.assertEqual(100, self.dwrap.pvid)
        self.assertEqual('https://9.1.2.3:12443/rest/api/uom/LogicalPartition/'
                         '0A68CFAB-F62B-46D4-A6A0-F4EBE0264AD5/'
                         'ClientNetworkAdapter/'
                         '6445b54b-b9dc-3bc2-b1d3-f8cc22ba95b8',
                         self.dwrap.href)
        self.assertEqual('U8246.L2C.0604C7A-V24-C32',
                         self.dwrap.loc_code)
        self.assertEqual([53, 54, 55], self.dwrap.tagged_vlans)
        self.assertTrue(self.dwrap.is_tagged_vlan_supported)
        self.assertEqual('https://9.1.2.3:12443/rest/api/uom/ManagedSystem/'
                         '726e9cb3-6576-3df5-ab60-40893d51d074/VirtualSwitch/'
                         '9e42d4a9-9725-3007-9932-d85374ebf5cf',
                         self.dwrap.vswitch_uri)
        self.assertEqual(0, self.dwrap.vswitch_id)
        self.assertEqual('VSITID', self.dwrap.vsi_type_id)
        self.assertEqual('77.99', self.dwrap.vsi_type_version)
        self.assertEqual('VSIMID', self.dwrap.vsi_type_manager_id)

    def test_tagged_vlan_modification(self):
        """Tests that the tagged vlans can be modified."""
        # Update via getter and Actionable List
        tags = self.dwrap.tagged_vlans
        tags.append(56)
        self.assertEqual(4, len(self.dwrap.tagged_vlans))
        tags.remove(56)
        self.assertEqual(3, len(self.dwrap.tagged_vlans))

        # Update via setter
        self.dwrap.tagged_vlans = [1, 2, 3]
        self.assertEqual([1, 2, 3], self.dwrap.tagged_vlans)
        self.dwrap.tagged_vlans = []
        self.assertEqual([], self.dwrap.tagged_vlans)
        self.dwrap.tagged_vlans = [53, 54, 55]

        # Try the tagged vlan support
        self.dwrap.is_tagged_vlan_supported = False
        self.assertFalse(self.dwrap.is_tagged_vlan_supported)
        self.dwrap.is_tagged_vlan_supported = True

    def test_mac_set(self):
        orig_mac = self.dwrap.mac
        mac = "AA:bb:CC:dd:ee:ff"
        self.dwrap.mac = mac
        self.assertEqual("AABBCCDDEEFF", self.dwrap.mac)
        self.dwrap.mac = orig_mac

    def test_get_slot(self):
        """Test getting the VirtualSlotID."""
        self.assertEqual(32, self.dwrap.slot)

    def test_get_mac(self):
        """Test that we can get the mac address."""
        self.assertEqual("FAD4433ED120", self.dwrap.mac)

    def test_pvid(self):
        """Test that the PVID returns properly."""
        self.assertEqual(100, self.dwrap.pvid)
        self.dwrap.pvid = 101
        self.assertEqual(101, self.dwrap.pvid)
        self.dwrap.pvid = 100

    def test_vswitch_uri(self):
        orig_uri = self.dwrap.vswitch_uri
        self.dwrap.vswitch_uri = 'test'
        self.assertEqual('test', self.dwrap.vswitch_uri)
        self.dwrap.vswitch_uri = orig_uri

    def test_wrapper_class(self):
        self.assertEqual(net.CNA.schema_type, 'ClientNetworkAdapter')
        self.assertEqual(net.CNA.schema_ns, pc.UOM_NS)
        self.assertTrue(net.CNA.has_metadata)
        self.assertEqual(net.CNA.default_attrib, pc.DEFAULT_SCHEMA_ATTR)

    def test_get_trunk_pri(self):
        """Test that we can get the trunk priority."""
        self.assertEqual(1, self.dwrap.trunk_pri)

    def test_set_trunk_pri(self):
        """Test that we can set the trunk priority."""
        self.assertEqual(1, self.dwrap.trunk_pri)
        self.dwrap._trunk_pri(2)
        self.assertEqual(2, self.dwrap.trunk_pri)

    def test_is_trunk(self):
        """Test that we can get if this adapter is a trunk."""
        self.assertTrue(self.dwrap.is_trunk)
        self.dwrap._trunk_pri(None)
        self.assertFalse(self.dwrap.is_trunk)

    def test_lpar_id(self):
        """Test that we can get the local partition id."""
        self.assertEqual(3, self.dwrap.lpar_id)

    def test_set_dev_name(self):
        """Test that we can set the device name."""
        self.assertEqual('Unknown', self.dwrap.dev_name)
        self.dwrap._dev_name('tap-01234')
        self.assertEqual('tap-01234', self.dwrap.dev_name)

    def test_enabled(self):
        """Test that we disable/enable."""
        self.assertTrue(self.dwrap.enabled)
        self.dwrap.enabled = False
        self.assertFalse(self.dwrap.enabled)
        self.dwrap.enabled = True
        self.assertTrue(self.dwrap.enabled)

if __name__ == "__main__":
    unittest.main()



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import random
import testtools

from pypowervm import adapter as adpt
from pypowervm import const as c
from pypowervm import exceptions as pvm_exc
from pypowervm.utils import retry as pvm_retry

called_count = 0


class TestRetry(testtools.TestCase):
    """Unit tests for pypowervm.util."""

    def test_retry(self):
        global called_count

        # Test normal call with return values
        @pvm_retry.retry(tries=4)
        def some_method(x, y):
            global called_count
            called_count += 1
            return x, y

        called_count = 0
        val = some_method(1, 2)
        self.assertEqual(val, (1, 2))
        self.assertEqual(called_count, 1)

        # Test with an unexpected exception
        class OurException(Exception):
            pass

        @pvm_retry.retry(tries=4)
        def except_method(x, y):
            global called_count
            called_count += 1
            raise OurException()

        called_count = 0
        self.assertRaises(OurException, except_method, 1, 2)
        self.assertEqual(called_count, 1)

        # Test retry with an http code
        @pvm_retry.retry(tries=4, http_codes=(c.HTTPStatus.ETAG_MISMATCH,))
        def http_except_method(x, y):
            global called_count
            called_count += 1
            resp = adpt.Response('reqmethod', 'reqpath',
                                 c.HTTPStatus.ETAG_MISMATCH, 'reason',
                                 'headers', None)
            http_exc = pvm_exc.HttpError(resp)
            raise http_exc

        called_count = 0
        self.assertRaises(pvm_exc.HttpError, http_except_method, 1, 2)
        self.assertEqual(called_count, 4)

        # Test retry with an test func and custom exception
        def cust_test(e, try_, tries, *args, **kwds):
            return try_ != 2

        @pvm_retry.retry(tries=10, test_func=cust_test,
                         limit_except=OurException())
        def func_except_method(x, y):
            global called_count
            called_count += 1
            resp = adpt.Response('reqmethod', 'reqpath',
                                 c.HTTPStatus.ETAG_MISMATCH, 'reason',
                                 'headers', None)
            http_exc = pvm_exc.HttpError(resp)
            raise http_exc

        called_count = 0
        # Should get back OurException after just 2 calls
        self.assertRaises(OurException, func_except_method, 1, 2)
        self.assertEqual(called_count, 2)

        # Test custom exceptions to retry
        @pvm_retry.retry(tries=3, retry_except=OurException)
        def func_except_method(x, y):
            global called_count
            called_count += 1
            raise OurException()

        called_count = 0
        # Should get back OurException after just 3 calls
        with self.assertLogs(pvm_retry.__name__, 'WARNING') as warn_logs:
            self.assertRaises(OurException, func_except_method, 1, 2)
            self.assertEqual(2, len(warn_logs.output))
        self.assertEqual(called_count, 3)

        # Test the response checking function
        def resp_chkr(resp, try_, tries, *args, **kwds):
            if try_ == 2:
                raise OurException()
            # Tell it to retry
            return True

        @pvm_retry.retry(tries=10, resp_checker=resp_chkr)
        def func_resp_chkr(x, y):
            global called_count
            called_count += 1
            return x, y

        called_count = 0
        # Should get back OurException after just 2 calls
        self.assertRaises(OurException, func_resp_chkr, 1, 2)
        self.assertEqual(called_count, 2)

    def test_retry_example(self):
        global called_count
        called_count = 0

        def _resp_checker(resp, try_, _tries, *args, **kwds):
            # If the VIOS is busy, then retry
            return resp == 'VIOS IS BUSY'

        @pvm_retry.retry(tries=4, http_codes=pvm_retry.DFT_RETRY_CODES,
                         resp_checker=_resp_checker)
        def _powervm_update(parm):
            global called_count
            called_count += 1
            if called_count == 1:
                # etag mismatch
                resp = adpt.Response('reqmethod', 'reqpath',
                                     c.HTTPStatus.ETAG_MISMATCH, 'reason',
                                     'headers')
                http_exc = pvm_exc.HttpError(resp)
                raise http_exc

            if called_count == 2:
                # Pretend we got a valid response, but the VIOS is busy
                return 'VIOS IS BUSY'

            if called_count == 3:
                # Pretend we got a good response
                return parm

            return None

        with self.assertLogs(pvm_retry.__name__, 'WARNING') as warn_logs:
            self.assertEqual(_powervm_update('Req'), 'Req')
            # only one warning (etag mismatch).  The 'VIOS IS BUSY' is
            # returned as OK by the _resp_checker, but doesn't do its own
            # logging
            self.assertEqual(1, len(warn_logs.output))
            self.assertEqual(called_count, 3)

    def test_retry_argmod(self):
        global called_count
        called_count = 0

        def argmod_func(this_try, max_tries, *args, **kwargs):
            argl = list(args)
            if this_try == 1:
                argl[0] += 1
            kwargs['five'] += ' bar'
            if this_try == 2:
                kwargs['seven'] = 7
            return argl, kwargs

        @pvm_retry.retry(argmod_func=argmod_func,
                         resp_checker=lambda *a, **kwa: True)
        def _func(one, two, three='four', five='six', seven=None):
            global called_count
            called_count += 1
            self.assertEqual(20, two)
            self.assertEqual('four', three)
            if called_count == 1:
                self.assertEqual(10, one)
                self.assertEqual('foo', five)
                self.assertIsNone(seven)
            else:
                self.assertEqual(11, one)
                if called_count == 2:
                    self.assertEqual('foo bar', five)
                elif called_count == 3:
                    self.assertEqual(7, seven)
                    self.assertEqual('foo bar bar', five)

        _func(10, 20, five='foo')
        self.assertEqual(3, called_count)

    def test_retry_refresh_wrapper(self):
        """Test @retry with the 'refresh_wrapper' argmod_func."""
        global called_count
        called_count = 0
        mock_wrapper = mock.Mock()
        mock_wrapper.refreshes = 0

        def _refresh(**kwargs):
            mock_wrapper.refreshes += 1
            self.assertIn('use_etag', kwargs)
            self.assertFalse(kwargs['use_etag'])
            return mock_wrapper
        mock_wrapper.refresh.side_effect = _refresh

        @pvm_retry.retry(argmod_func=pvm_retry.refresh_wrapper,
                         resp_checker=lambda *a, **k: True)
        def _func(wrapper, arg1, arg2, kw0=None, kw1=None):
            global called_count
            self.assertEqual(called_count, wrapper.refreshes)
            # Ensure the other args didn't change
            self.assertEqual('a1', arg1)
            self.assertEqual('a2', arg2)
            self.assertEqual('k0', kw0)
            self.assertEqual('k1', kw1)
            called_count += 1
        _func(mock_wrapper, 'a1', 'a2', kw0='k0', kw1='k1')
        # Three calls (overall attempts)
        self.assertEqual(3, called_count)
        # ...equals two refreshes
        self.assertEqual(2, mock_wrapper.refreshes)

    @mock.patch('time.sleep')
    def test_stepped_delay(self, mock_sleep):
        # Last set of delays should hit the cap.
        delays = [0, .5, 2.0, 6.5, 20.0, 30.0, 30.0]
        for i in range(1, 7):
            pvm_retry.STEPPED_DELAY(i, 7)
            mock_sleep.assert_called_once_with(delays[i-1])
            mock_sleep.reset_mock()

    @mock.patch('time.sleep')
    def test_random_delay(self, mock_sleep):
        """Test gen_random_delay."""
        def _validate_range(start, end):
            # Sleep was called once.
            self.assertEqual(1, mock_sleep.call_count)
            args, kwargs = mock_sleep.call_args
            # Called with one arg
            self.assertEqual(1, len(args))
            # ...and no kwargs
            self.assertEqual({}, kwargs)
            # Extract the arg
            slept_with = args[0]
            # It should be at least 'start'
            self.assertGreaterEqual(slept_with, start)
            # ...and at most 'end'.
            self.assertLessEqual(slept_with, end)
            mock_sleep.reset_mock()
        # Defaults
        pvm_retry.gen_random_delay()(1, 1)
        _validate_range(0, 10)
        # Use a randomizer to test a randomizer.  It's... poetry.
        min_s = random.random() * random.randint(1, 100)
        max_s = min_s * random.randint(1, 100)
        pvm_retry.gen_random_delay(min_s=min_s, max_s=max_s)(1, 1)
        _validate_range(min_s, max_s)

    def _validate_stepped_random_range(self, mock_sleep, attempt, start, end,
                                       max_attempts=6):
        pvm_retry.STEPPED_RANDOM_DELAY(attempt, max_attempts)
        # Sleep was called once.
        self.assertEqual(1, mock_sleep.call_count)
        args, kwargs = mock_sleep.call_args
        # Called with one arg
        self.assertEqual(1, len(args))
        # ...and no kwargs
        self.assertEqual({}, kwargs)
        # Extract the arg
        slept_with = args[0]
        # It should be at least 'start'
        self.assertGreaterEqual(slept_with, start)
        # ...and at most 'end'.
        self.assertLessEqual(slept_with, end)
        mock_sleep.reset_mock()

    @mock.patch('time.sleep')
    def test_stepped_random_delay(self, mock_sleep):
        """Test STEPPED_RANDOM_DELAY."""

        # These ranges from RANDOM_DELAY_STEPS
        self._validate_stepped_random_range(mock_sleep, 1, 0, 0)
        self._validate_stepped_random_range(mock_sleep, 2, 0, 1)
        self._validate_stepped_random_range(mock_sleep, 3, 0.5, 4)
        self._validate_stepped_random_range(mock_sleep, 4, 2, 13)
        self._validate_stepped_random_range(mock_sleep, 5, 6.5, 30)

        # Use a range beyond the max attempts.  Make sure it does not fail.
        for att in (6, 7, 8, 9, 10):
            self._validate_stepped_random_range(mock_sleep, att, 0, 60)

    @mock.patch('time.sleep')
    def test_stepped_random_delay_large(self, mock_sleep):
        """Test STEPPED_RANDOM_DELAY with a very large range."""

        max_attempts = 60

        # These ranges from RANDOM_DELAY_STEPS
        for i in range(1, 10):
            self._validate_stepped_random_range(mock_sleep, i, 0, 0,
                                                max_attempts=max_attempts)
        for i in range(11, 20):
            self._validate_stepped_random_range(mock_sleep, i, 0, 1,
                                                max_attempts=max_attempts)
        for i in range(21, 30):
            self._validate_stepped_random_range(mock_sleep, i, 0.5, 4,
                                                max_attempts=max_attempts)
        for i in range(31, 40):
            self._validate_stepped_random_range(mock_sleep, i, 2, 13,
                                                max_attempts=max_attempts)
        for i in range(41, 50):
            self._validate_stepped_random_range(mock_sleep, i, 6.5, 30,
                                                max_attempts=max_attempts)
        for i in range(51, 60):
            self._validate_stepped_random_range(mock_sleep, i, 0, 60,
                                                max_attempts=max_attempts)

    @mock.patch('time.sleep')
    def test_stepped_random_delay_small(self, mock_sleep):
        """Test STEPPED_RANDOM_DELAY with a very small range.

        Uses a 'max_attempts' whose range is less than the size of
        RANDOM_DELAY_STEPS
        """
        # Run this a few times so as to make sure the ranges are fully
        # adhered to.
        i = 0
        while i < 50:
            i += 1
            self._validate_stepped_random_range(mock_sleep, 1, 0, 0,
                                                max_attempts=3)
            self._validate_stepped_random_range(mock_sleep, 2, .5, 4,
                                                max_attempts=3)
            self._validate_stepped_random_range(mock_sleep, 3, 6.5, 60,
                                                max_attempts=3)



# Copyright 2017 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import testtools

import pypowervm.exceptions as exc
import pypowervm.tasks.power_opts as popts
import pypowervm.wrappers.base_partition as bp


class TestPowerOpts(testtools.TestCase):
    def _test_enum(self, enum):
        """Validate that an enum class has a KEY and proper ALL_VALUES.

        :param enum: Enumeration class.
        """
        # Get the public symbols in the enum
        syms = {sym for sym in dir(enum) if not sym.startswith('_')}
        # Must have a KEY
        self.assertIn('KEY', syms)
        self.assertIsNotNone(getattr(enum, 'KEY'))
        syms.remove('KEY')
        # Must have ALL_VALUES
        self.assertIn('ALL_VALUES', syms)
        syms.remove('ALL_VALUES')
        # ALL_VALUES must include all the values that aren't KEY or ALL_VALUES
        self.assertEqual({getattr(enum, sym) for sym in syms},
                         set(enum.ALL_VALUES))

    def test_enums(self):
        self._test_enum(popts.IPLSrc)
        self._test_enum(popts.BootMode)
        self._test_enum(popts.KeylockPos)
        self._test_enum(popts.IBMiOperationType)
        self._test_enum(popts.PowerOffOperation)

    def test_remove_optical(self):
        knm = popts.RemoveOptical.KEY_NAME
        ktm = popts.RemoveOptical.KEY_TIME
        # Default time
        self.assertEqual({knm: 'name', ktm: 0},
                         popts.RemoveOptical.bld_map('name'))
        # Explicit time
        self.assertEqual({knm: 'name', ktm: 10},
                         popts.RemoveOptical.bld_map('name', time=10))

    def test_power_on_opts(self):
        # Default init
        poo = popts.PowerOnOpts()
        self.assertEqual('PowerOn()', str(poo))
        self.assertEqual('PowerOn', poo.JOB_SUFFIX)
        # Legacy add_parms init
        poo = popts.PowerOnOpts(legacy_add_parms=dict(foo=1, bar=2))
        self.assertEqual('PowerOn(bar=2, foo=1)', str(poo))
        # Carry those additional params forward to make sure they don't vanish
        # Enum validation
        for meth in ('bootmode', 'keylock_pos', 'ibmi_ipl_source',
                     'ibmi_op_type'):
            self.assertRaises(exc.InvalidEnumValue, getattr(poo, meth), 'foo')
        # Set specific (valid) values
        # Setter method returns the instance
        self.assertIs(poo, poo.bootmode(popts.BootMode.NORM))
        self.assertEqual('PowerOn(bar=2, bootmode=norm, foo=1)', str(poo))
        self.assertIs(poo, poo.keylock_pos(popts.KeylockPos.MANUAL))
        self.assertEqual(
            'PowerOn(bar=2, bootmode=norm, foo=1, keylock=manual)', str(poo))
        self.assertIs(poo, poo.bootstring('canvas cord with aglet'))
        self.assertEqual(
            'PowerOn(bar=2, bootmode=norm, bootstring=canvas cord with aglet, '
            'foo=1, keylock=manual)', str(poo))
        # Make sure overwrite works
        self.assertIs(poo, poo.bootstring('sturdy shoelace'))
        self.assertEqual(
            'PowerOn(bar=2, bootmode=norm, bootstring=sturdy shoelace, foo=1, '
            'keylock=manual)', str(poo))
        self.assertIs(poo, poo.force())
        self.assertEqual(
            'PowerOn(bar=2, bootmode=norm, bootstring=sturdy shoelace, foo=1, '
            'force=true, keylock=manual)', str(poo))
        # Turning off force gets rid of the key
        self.assertIs(poo, poo.force(value=False))
        self.assertEqual(
            'PowerOn(bar=2, bootmode=norm, bootstring=sturdy shoelace, foo=1, '
            'keylock=manual)', str(poo))
        # Remove optical with default time
        self.assertIs(poo, poo.remove_optical('vopt'))
        self.assertEqual(
            'PowerOn(bar=2, bootmode=norm, bootstring=sturdy shoelace, foo=1, '
            'keylock=manual, remove_optical_name=vopt, remove_optical_time=0)',
            str(poo))
        # Remove optical with explicit time.  Values are overwritten.
        self.assertIs(poo, poo.remove_optical('VOPT', time=5))
        self.assertEqual(
            'PowerOn(bar=2, bootmode=norm, bootstring=sturdy shoelace, foo=1, '
            'keylock=manual, remove_optical_name=VOPT, remove_optical_time=5)',
            str(poo))
        self.assertIs(poo, poo.ibmi_ipl_source(popts.IPLSrc.A))
        self.assertEqual(
            'PowerOn(bar=2, bootmode=norm, bootstring=sturdy shoelace, foo=1, '
            'iIPLsource=a, keylock=manual, remove_optical_name=VOPT, '
            'remove_optical_time=5)', str(poo))
        self.assertIs(poo, poo.ibmi_op_type(popts.IBMiOperationType.NETBOOT))
        self.assertEqual(
            'PowerOn(OperationType=netboot, bar=2, bootmode=norm, '
            'bootstring=sturdy shoelace, foo=1, iIPLsource=a, keylock=manual, '
            'remove_optical_name=VOPT, remove_optical_time=5)', str(poo))
        # Netboot params.
        poo = popts.PowerOnOpts().ibmi_netboot_params(
            'ip', 'serverip', 'gateway', 'serverdir')
        self.assertEqual(
            'PowerOn(Gateway=gateway, IBMiImageServerDirectory=serverdir, '
            'IPAddress=ip, ServerIPAddress=serverip)', str(poo))
        # Optional netboot params, and overwrites
        self.assertIs(poo, poo.ibmi_netboot_params(
            'IP', 'ServerIP', 'Gateway', 'ServerDir', vlanid=2, mtu='mtu',
            duplex='duplex', connspeed=100, subnet='subnet'))
        self.assertEqual(
            'PowerOn(ConnectionSpeed=100, DuplexMode=duplex, Gateway=Gateway, '
            'IBMiImageServerDirectory=ServerDir, IPAddress=IP, '
            'MaximumTransmissionUnit=mtu, ServerIPAddress=ServerIP, '
            'SubnetMask=subnet, VLANID=2)', str(poo))

    def test_power_off_opts(self):
        # Can OS shutdown?
        ltyp = bp.LPARType
        rmcs = bp.RMCState
        for env, rmc, exp in ((ltyp.AIXLINUX, rmcs.ACTIVE, True),
                              (ltyp.AIXLINUX, rmcs.BUSY, False),
                              (ltyp.AIXLINUX, rmcs.INACTIVE, False),
                              (ltyp.OS400, rmcs.ACTIVE, True),
                              (ltyp.OS400, rmcs.BUSY, True),
                              (ltyp.OS400, rmcs.INACTIVE, True),
                              (ltyp.VIOS, rmcs.ACTIVE, True),
                              (ltyp.VIOS, rmcs.BUSY, False),
                              (ltyp.VIOS, rmcs.INACTIVE, False)):
            self.assertEqual(exp, popts.PowerOffOpts.can_os_shutdown(
                mock.Mock(env=env, rmc_state=rmc)))
        # Default init
        poo = popts.PowerOffOpts()
        self.assertEqual('PowerOff()', str(poo))
        self.assertEqual('PowerOff', poo.JOB_SUFFIX)
        # Legacy add_parms init
        poo = popts.PowerOffOpts(legacy_add_parms=dict(foo=1, bar=2))
        self.assertEqual('PowerOff(bar=2, foo=1)', str(poo))
        self.assertFalse(poo.is_immediate)
        self.assertFalse(poo.is_restart)
        self.assertFalse(poo.is_os)
        # Carry those additional params forward to make sure they don't vanish
        # Immediate
        self.assertIs(poo, poo.immediate())
        self.assertEqual('PowerOff(bar=2, foo=1, immediate=true)', str(poo))
        self.assertTrue(poo.is_immediate)
        self.assertFalse(poo.is_restart)
        self.assertFalse(poo.is_os)
        # Restart
        self.assertIs(poo, poo.restart())
        self.assertEqual(
            'PowerOff(bar=2, foo=1, immediate=true, restart=true)', str(poo))
        self.assertTrue(poo.is_immediate)
        self.assertTrue(poo.is_restart)
        self.assertFalse(poo.is_os)
        # Operation
        self.assertIs(poo, poo.operation(popts.PowerOffOperation.DUMPRESTART))
        self.assertEqual(
            'PowerOff(bar=2, foo=1, immediate=true, operation=dumprestart, '
            'restart=true)', str(poo))
        self.assertTrue(poo.is_immediate)
        self.assertTrue(poo.is_restart)
        self.assertFalse(poo.is_os)
        # OS shutdown
        self.assertIs(poo, poo.operation(popts.PowerOffOperation.OS))
        self.assertEqual(
            'PowerOff(bar=2, foo=1, immediate=true, operation=osshutdown, '
            'restart=true)', str(poo))
        self.assertTrue(poo.is_immediate)
        self.assertTrue(poo.is_restart)
        self.assertTrue(poo.is_os)
        # Booleans can be shut off
        self.assertIs(poo, poo.immediate(value=False))
        self.assertEqual('PowerOff(bar=2, foo=1, operation=osshutdown, '
                         'restart=true)', str(poo))
        self.assertFalse(poo.is_immediate)
        self.assertTrue(poo.is_restart)
        self.assertTrue(poo.is_os)
        self.assertIs(poo, poo.restart(value=False))
        self.assertEqual('PowerOff(bar=2, foo=1, operation=osshutdown)',
                         str(poo))
        self.assertFalse(poo.is_immediate)
        self.assertFalse(poo.is_restart)
        self.assertTrue(poo.is_os)
        # "Smart" methods.  Make sure restart is preserved every time we change
        poo.restart()
        # OS immediate
        self.assertIs(poo, poo.os_immediate())
        self.assertEqual('PowerOff(bar=2, foo=1, immediate=true, '
                         'operation=osshutdown, restart=true)', str(poo))
        self.assertTrue(poo.is_immediate)
        self.assertTrue(poo.is_restart)
        self.assertTrue(poo.is_os)
        # OS normal (wipes out immediate)
        self.assertIs(poo, poo.os_normal())
        self.assertEqual('PowerOff(bar=2, foo=1, operation=osshutdown, '
                         'restart=true)', str(poo))
        self.assertFalse(poo.is_immediate)
        self.assertTrue(poo.is_restart)
        self.assertTrue(poo.is_os)
        # VSP hard
        self.assertIs(poo, poo.vsp_hard())
        self.assertEqual('PowerOff(bar=2, foo=1, immediate=true, '
                         'operation=shutdown, restart=true)', str(poo))
        self.assertTrue(poo.is_immediate)
        self.assertTrue(poo.is_restart)
        self.assertFalse(poo.is_os)
        # VSP normal (wipes out immediate)
        self.assertIs(poo, poo.vsp_normal())
        self.assertEqual('PowerOff(bar=2, foo=1, operation=shutdown, '
                         'restart=true)', str(poo))
        self.assertFalse(poo.is_immediate)
        self.assertTrue(poo.is_restart)
        self.assertFalse(poo.is_os)
        # Soft detect
        part = mock.Mock(env=ltyp.AIXLINUX, rmc_state=rmcs.ACTIVE)
        self.assertIs(poo, poo.soft_detect(part))
        self.assertTrue(poo.is_os)
        self.assertTrue(poo.is_immediate)
        self.assertTrue(poo.is_restart)
        # Explicit normal shutdown
        self.assertIs(poo, poo.soft_detect(part, immed_if_os=False))
        self.assertTrue(poo.is_os)
        self.assertFalse(poo.is_immediate)
        self.assertTrue(poo.is_restart)
        # Explicit immediate OS shutdown
        self.assertIs(poo, poo.soft_detect(part, immed_if_os=True))
        self.assertTrue(poo.is_os)
        self.assertTrue(poo.is_immediate)
        self.assertTrue(poo.is_restart)
        # Can't OS shutdown
        part = mock.Mock(env=ltyp.VIOS, rmc_state=rmcs.BUSY)
        self.assertIs(poo, poo.soft_detect(part))
        self.assertFalse(poo.is_os)
        self.assertFalse(poo.is_immediate)
        self.assertTrue(poo.is_restart)
        # immed_if_os ignored
        self.assertIs(poo, poo.soft_detect(part, immed_if_os=True))
        self.assertFalse(poo.is_os)
        self.assertFalse(poo.is_immediate)
        self.assertTrue(poo.is_restart)
        self.assertIs(poo, poo.soft_detect(part, immed_if_os=False))
        self.assertFalse(poo.is_os)
        self.assertFalse(poo.is_immediate)
        self.assertTrue(poo.is_restart)
        self.assertEqual('PowerOff(bar=2, foo=1, operation=shutdown, '
                         'restart=true)', str(poo))
        # IBMi defaults to OS normal
        part = mock.Mock(env=ltyp.OS400, rmc_state=rmcs.INACTIVE)
        self.assertIs(poo, poo.soft_detect(part))
        self.assertTrue(poo.is_os)
        self.assertFalse(poo.is_immediate)
        self.assertTrue(poo.is_restart)
        # Explicit immediate
        self.assertIs(poo, poo.soft_detect(part, immed_if_os=True))
        self.assertTrue(poo.is_os)
        self.assertTrue(poo.is_immediate)
        self.assertTrue(poo.is_restart)



# Copyright 2014, 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import six

if six.PY2:
    import __builtin__ as builtins
elif six.PY3:
    import builtins

import unittest

from pypowervm import const
from pypowervm import util

dummyuuid1 = "abcdef01-2345-2345-2345-67890abcdef0"
dummyuuid2 = "67890abc-5432-5432-5432-def0abcdef01"


class TestUtil(unittest.TestCase):
    """Unit tests for pypowervm.util."""

    def test_convert_bytes_to_gb(self):
        # A round 1 GB
        test = util.convert_bytes_to_gb(1024 * 1024 * 1024)
        self.assertEqual(1.0, test)

        # A single MB
        test = util.convert_bytes_to_gb(1024 * 1024.0)
        self.assertEqual(0.0009765625, test)

        # A single byte - should be the low Value
        self.assertEqual(.0001, util.convert_bytes_to_gb(1))

        # Try changing the low value
        self.assertEqual(.0005, util.convert_bytes_to_gb(1, .0005))

        # Round up
        self.assertEqual(1.15, util.convert_bytes_to_gb(1224067890, dp=2))

        # Low value still honors dp
        self.assertEqual(0.01, util.convert_bytes_to_gb(1, dp=2))

    def test_round_gb_size_up(self):
        self.assertEqual(12.35, util.round_gb_size_up(12.34000000001))
        self.assertEqual(12.34000000001, util.round_gb_size_up(12.34000000001,
                                                               dp=11))
        self.assertEqual(1048576, util.round_gb_size_up(1048576.0, dp=0))
        self.assertEqual(1048576, util.round_gb_size_up(1048575.1, dp=0))
        self.assertEqual(1048576, util.round_gb_size_up(1048576, dp=0))
        self.assertEqual(1048600, util.round_gb_size_up(1048576.1234, dp=-2))

    def test_sanitize_bool_for_api(self):
        self.assertEqual('true', util.sanitize_bool_for_api(True))
        self.assertEqual('false', util.sanitize_bool_for_api(False))
        self.assertEqual('true', util.sanitize_bool_for_api('True'))
        self.assertEqual('false', util.sanitize_bool_for_api('False'))

    def test_find_wrapper(self):
        wrap1 = mock.MagicMock()
        wrap1.uuid = 'a'
        wrap2 = mock.MagicMock()
        wrap2.uuid = 'b'
        wraps = [wrap1, wrap2]

        self.assertEqual(wrap1, util.find_wrapper(wraps, 'a'))
        self.assertEqual(wrap2, util.find_wrapper(wraps, 'b'))
        self.assertIsNone(util.find_wrapper(wraps, 'c'))

    def test_dice_href(self):
        href = 'https://server:1234/rest/api/uom/Obj/UUID//?group=One,Two#frag'
        self.assertEqual(util.dice_href(href),
                         '/rest/api/uom/Obj/UUID?group=One,Two#frag')
        self.assertEqual(util.dice_href(href, include_query=True),
                         '/rest/api/uom/Obj/UUID?group=One,Two#frag')
        self.assertEqual(util.dice_href(href, include_fragment=False),
                         '/rest/api/uom/Obj/UUID?group=One,Two')
        self.assertEqual(util.dice_href(href, include_query=False),
                         '/rest/api/uom/Obj/UUID#frag')
        self.assertEqual(util.dice_href(href, include_fragment=True),
                         '/rest/api/uom/Obj/UUID?group=One,Two#frag')
        self.assertEqual(util.dice_href(href, include_query=False,
                                        include_fragment=True),
                         '/rest/api/uom/Obj/UUID#frag')
        self.assertEqual(util.dice_href(href, include_scheme_netloc=True,
                                        include_query=False,
                                        include_fragment=False),
                         'https://server:1234/rest/api/uom/Obj/UUID')

    def test_get_req_path_uuid_and_is_instance_path(self):
        # Fail: no '/'
        path = dummyuuid1
        self.assertIsNone(util.get_req_path_uuid(path))
        self.assertRaises(IndexError, util.is_instance_path, path)
        path = '/' + dummyuuid1
        self.assertEqual(dummyuuid1, util.get_req_path_uuid(path))
        self.assertTrue(util.is_instance_path(path))
        path = 'https://server:1234/rest/api/uom/Obj/' + dummyuuid1
        self.assertEqual(dummyuuid1, util.get_req_path_uuid(path))
        self.assertTrue(util.is_instance_path(path))
        # Fail: last path element is not a UUID
        path = 'https://server:1234/rest/api/uom/Obj/' + dummyuuid1 + '/Child'
        self.assertIsNone(util.get_req_path_uuid(path))
        self.assertFalse(util.is_instance_path(path))
        # Fail: last path element is not quiiiite a UUID
        path = 'https://server:1234/rest/api/uom/Obj/' + dummyuuid1[1:]
        self.assertIsNone(util.get_req_path_uuid(path))
        self.assertFalse(util.is_instance_path(path))
        # Ignore query/fragment
        path = ('https://server:1234/rest/api/uom/Obj/' + dummyuuid1 +
                '?group=One,Two#frag')
        self.assertEqual(dummyuuid1, util.get_req_path_uuid(path))
        self.assertTrue(util.is_instance_path(path))
        # Fail: last path element (having removed query/fragment) is not a UUID
        path = ('https://server:1234/rest/api/uom/Obj/' + dummyuuid1 +
                '/Child?group=One,Two#frag')
        self.assertIsNone(util.get_req_path_uuid(path))
        self.assertFalse(util.is_instance_path(path))
        # Default case conversion
        path = 'https://server:1234/rest/api/uom/Obj/' + dummyuuid1.upper()
        self.assertEqual(dummyuuid1, util.get_req_path_uuid(path))
        self.assertEqual(dummyuuid1, util.get_req_path_uuid(
            path, preserve_case=False))
        self.assertTrue(util.is_instance_path(path))
        # Force no case conversion
        self.assertEqual(dummyuuid1.upper(), util.get_req_path_uuid(
            path, preserve_case=True))
        # Child URI gets child UUID by default
        path = ('https://server:1234/rest/api/uom/Obj/' + dummyuuid1 +
                '/Child/' + dummyuuid2)
        self.assertEqual(dummyuuid2, util.get_req_path_uuid(path))
        self.assertTrue(util.is_instance_path(path))
        # Get root UUID from child URI
        path = ('https://server:1234/rest/api/uom/Obj/' + dummyuuid1 +
                '/Child/' + dummyuuid2)
        self.assertEqual(dummyuuid1, util.get_req_path_uuid(path, root=True))
        self.assertTrue(util.is_instance_path(path))
        # root=True redundant on a root path
        path = '/' + dummyuuid1
        self.assertEqual(dummyuuid1, util.get_req_path_uuid(path, root=True))
        path = 'https://server:1234/rest/api/uom/Obj/' + dummyuuid1
        self.assertEqual(dummyuuid1, util.get_req_path_uuid(path, root=True))

    def test_extend_basepath(self):
        ext = '/foo'
        # Various forms without query params or fragments
        for path in (dummyuuid1, '/' + dummyuuid1,
                     'https://server:1234/rest/api/uom/Obj/' + dummyuuid1,
                     'https://server:1234/rest/api/uom/Obj/' + dummyuuid1 +
                     '/Child'):
            self.assertEqual(path + ext, util.extend_basepath(path, ext))

        basepath = 'https://server:1234/rest/api/uom/Obj/' + dummyuuid1
        qp = '?foo=bar,baz&blah=123'
        frag = '#frag'
        # Query params
        self.assertEqual(basepath + ext + qp,
                         util.extend_basepath(basepath + qp, ext))
        # Fragment
        self.assertEqual(basepath + ext + frag,
                         util.extend_basepath(basepath + frag, ext))
        # Query params & fragment
        self.assertEqual(basepath + ext + qp + frag,
                         util.extend_basepath(basepath + qp + frag, ext))

    def test_sanitize_file_name_for_api(self):
        allc = ''.join(map(chr, range(256)))
        self.assertEqual('foo', util.sanitize_file_name_for_api('foo'))
        self.assertEqual(
            'config_foo.iso', util.sanitize_file_name_for_api(
                'foo', prefix='config_', suffix='.iso'))
        self.assertEqual(
            '______________________________________________._0123456789_______'
            'ABCDEFGHIJKLMN',
            util.sanitize_file_name_for_api(allc))
        self.assertEqual(
            'OPQRSTUVWXYZ______abcdefghijklmnopqrstuvwxyz_____________________'
            '______________',
            util.sanitize_file_name_for_api(allc[79:])
        )
        self.assertEqual(
            '_________________________________________________________________'
            '______________',
            util.sanitize_file_name_for_api(allc[158:])
        )
        self.assertEqual('___________________',
                         util.sanitize_file_name_for_api(allc[237:]))
        self.assertEqual(
            (dummyuuid1 + dummyuuid2[:7] + dummyuuid1).replace('-', '_'),
            util.sanitize_file_name_for_api(
                dummyuuid2, prefix=dummyuuid1, suffix=dummyuuid1))
        self.assertEqual('I____________',
                         util.sanitize_file_name_for_api(
                             u'I \u611B \u01A4\u0177\u03C1\uFF4F\u05E9\u5DF3'
                             u'\u5C3A\uFF56\uFF4D'))
        self.assertRaises(ValueError, util.sanitize_file_name_for_api, allc,
                          prefix=allc, suffix=allc)
        self.assertRaises(ValueError, util.sanitize_file_name_for_api, '')
        # Non-default max_len values
        self.assertEqual('abcdefghijklmno', util.sanitize_file_name_for_api(
            'abcdefghijklmnopqrstuvwxyz', max_len=const.MaxLen.VDISK_NAME))
        self.assertEqual(
            'abcdefghijklmnopqrstuvwxyz0123456789A',
            util.sanitize_file_name_for_api(
                'abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNO',
                max_len=const.MaxLen.VOPT_NAME))

    def test_sanitize_partition_name_for_api(self):
        allc = ''.join(map(chr, range(256)))
        self.assertEqual('foo', util.sanitize_partition_name_for_api('foo'))
        self.assertEqual('_______________________________',
                         util.sanitize_partition_name_for_api(allc))
        self.assertEqual('_ !_#_%_____+,-./0123456789:;_=',
                         util.sanitize_partition_name_for_api(allc[31:]))
        self.assertEqual('__@ABCDEFGHIJKLMNOPQRSTUVWXYZ__',
                         util.sanitize_partition_name_for_api(allc[62:]))
        self.assertEqual('_^__abcdefghijklmnopqrstuvwxyz{',
                         util.sanitize_partition_name_for_api(allc[93:]))
        self.assertEqual('_}_____________________________',
                         util.sanitize_partition_name_for_api(allc[124:]))
        for start in (155, 186, 217):
            self.assertEqual(
                '_______________________________',
                util.sanitize_partition_name_for_api(allc[start:]))
        self.assertEqual('________',
                         util.sanitize_partition_name_for_api(allc[248:]))
        self.assertEqual('I _ _________',
                         util.sanitize_partition_name_for_api(
                             u'I \u611B \u01A4\u0177\u03C1\uFF4F\u05E9\u5DF3'
                             u'\u5C3A\uFF56\uFF4D'))

        self.assertRaises(ValueError, util.sanitize_partition_name_for_api,
                          allc, trunc_ok=False)
        self.assertRaises(ValueError, util.sanitize_partition_name_for_api, '')
        self.assertRaises(ValueError, util.sanitize_partition_name_for_api,
                          None)

    # Tests for check_and_apply_xag covered by
    # test_adapter.TestAdapter.test_extended_path

    def test_part_id_by_loc_code(self):
        test_loc = 'U8247.22L.2125D6A-V2-C3'
        fail_loc = 'abc1234'
        self.assertEqual(util.part_id_by_loc_code(test_loc), 2)
        self.assertIsNone(util.part_id_by_loc_code(fail_loc))

    def test_xag_attrs(self):
        base = const.DEFAULT_SCHEMA_ATTR
        self.assertEqual(dict(base), util.xag_attrs(''))
        self.assertEqual(dict(base), util.xag_attrs(None))
        self.assertEqual(dict(base, group='foo'), util.xag_attrs('foo'))
        # Test other bases
        self.assertEqual(dict(one=2), util.xag_attrs(None, base=dict(one=2)))
        self.assertEqual(dict(one=2, group='foo'),
                         util.xag_attrs('foo', base=dict(one=2)))

    @mock.patch.object(builtins, 'open')
    def test_my_partition_id(self, m_open):
        """Test my_partition_id."""
        def rit():
            for line in ('foo=bar\n', 'partition_id=1234\n', '\n', 'a=b\n'):
                yield line
        m_open.return_value.__enter__.return_value.__iter__.side_effect = rit
        self.assertEqual(1234, util.my_partition_id())

    def test_parent_spec(self):
        """Test parent_spec."""
        # All params are None (ROOT request)
        self.assertEqual((None, None), util.parent_spec(None, None, None))
        # Get values from parent
        parent = mock.Mock(schema_type='schema_type', uuid='uuid')
        self.assertEqual(('schema_type', 'uuid'), util.parent_spec(
            parent, None, None))
        # Parent overrides parent_type/parent_uuid
        self.assertEqual(('schema_type', 'uuid'), util.parent_spec(
            parent, 'something', 'else'))
        # ValueError if type xor uuid specified
        self.assertRaises(ValueError, util.parent_spec, None, 'one', None)
        self.assertRaises(ValueError, util.parent_spec, None, None, 'two')
        # Non-wrapper, non-string parent type raises ValueError
        self.assertRaises(ValueError, util.parent_spec, None, 42, 'foo')
        # parent_type can be wrapper or string
        self.assertEqual(('schema_type', 'uuid2'), util.parent_spec(
            None, parent, 'uuid2'))
        self.assertEqual(('schema_type2', 'uuid2'), util.parent_spec(
            None, 'schema_type2', 'uuid2'))

    def test_retry_io_command(self):
        class MyOSError(OSError):
            def __init__(self, errno):
                super(MyOSError, self).__init__()
                self.errno = errno

        class MyIOError(IOError):
            def __init__(self, errno):
                super(MyIOError, self).__init__()
                self.errno = errno

        class MyValError(ValueError):
            def __init__(self, errno):
                super(MyValError, self).__init__()
                self.errno = errno

        func = mock.Mock()
        mock_os_intr = MyOSError(4)
        mock_io_intr = MyIOError(4)
        mock_val_intr = MyValError(4)
        mock_os_hup = MyOSError(1)
        mock_io_hup = MyIOError(1)
        func.side_effect = [mock_os_intr, mock_io_intr, mock_val_intr]
        self.assertRaises(MyValError, util.retry_io_command, func)
        self.assertEqual(3, func.call_count)
        func.reset_mock()
        func.side_effect = mock_os_hup
        self.assertRaises(MyOSError, util.retry_io_command, func, 1, 'a')
        func.assert_called_once_with(1, 'a')
        func.reset_mock()
        func.side_effect = mock_io_hup
        self.assertRaises(MyIOError, util.retry_io_command, func)
        func.assert_called_once_with()


class TestAllowedList(unittest.TestCase):
    def test_all_none(self):
        for cls in (util.VLANList, util.MACList):
            for val in ('ALL', 'NONE'):
                self.assertEqual(val, cls.unmarshal(val))
            for val in ('ALL', 'NONE', 'all', 'none', 'aLl', 'nOnE'):
                self.assertEqual(val.upper(), cls.marshal(val))
                self.assertEqual(val.upper(), cls.const_or_list(val))
                self.assertEqual(val.upper(), cls.marshal([val]))
                self.assertEqual(val.upper(), cls.const_or_list([val]))

    def test_unmarshal(self):
        # Test VLAN lists
        self.assertEqual([1, 2], util.VLANList.unmarshal('1 2'))
        self.assertEqual([0], util.VLANList.unmarshal('0'))
        self.assertEqual([5, 6, 2230, 3340],
                         util.VLANList.unmarshal('5 6 2230 3340'))

        # Test MAC lists
        self.assertEqual(['AB12CD34EF56', '12AB34CD56EF'],
                         util.MACList.unmarshal('AB12CD34EF56 12AB34CD56EF'))
        self.assertEqual(['AB12CD34EF56'],
                         util.MACList.unmarshal('AB12CD34EF56'))

    def test_marshal(self):
        # Test VLAN lists
        self.assertEqual('1 2', util.VLANList.marshal([1, 2]))
        self.assertEqual('0', util.VLANList.marshal([0]))
        self.assertEqual('5 6 2230 3340',
                         util.VLANList.marshal([5, 6, '2230', 3340]))

        # Test MAC lists
        self.assertEqual('AB12CD34EF56 12AB34CD56EF', util.MACList.marshal(
            ['aB:12:Cd:34:eF:56', '12Ab34cD56Ef']))
        self.assertEqual('AB12CD34EF56', util.MACList.marshal(
            ['Ab:12:cD:34:Ef:56']))

        # Test error cases
        for cls in (util.VLANList, util.MACList):
            self.assertRaises(ValueError, cls.marshal, None)
            self.assertRaises(ValueError, cls.marshal, '')
            self.assertRaises(ValueError, cls.marshal, ' ')
            self.assertRaises(ValueError, cls.marshal, 'bogus')

    def test_const_or_list(self):
        # Test VLAN lists
        for l2t in ([1, 2], [0], [5, 6, 2230, 3340]):
            self.assertEqual(l2t, util.VLANList.const_or_list(l2t))

        # Test MAC lists
        self.assertEqual(['AB12CD34EF56', '12AB34CD56EF'],
                         util.MACList.const_or_list(
                             ['aB:12:Cd:34:eF:56', '12Ab34cD56Ef']))
        self.assertEqual(['AB12CD34EF56'], util.MACList.const_or_list(
            ['Ab:12:cD:34:Ef:56']))

        # Test error cases
        for cls in (util.VLANList, util.MACList):
            for meth in (cls.marshal, cls.const_or_list):
                self.assertRaises(ValueError, meth, None)
                self.assertRaises(ValueError, meth, '')
                self.assertRaises(ValueError, meth, ' ')
                self.assertRaises(ValueError, meth, 'bogus')
        self.assertRaises(ValueError, util.VLANList.marshal, ['1', 'NaN', 2])
        self.assertRaises(ValueError, util.VLANList.const_or_list, ['1', 'NaN',
                                                                    2])



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from pypowervm.tests.test_utils import test_wrapper_abc
from pypowervm.wrappers import enterprise_pool as ep
from pypowervm.wrappers import entry_wrapper


class TestPoolFile(test_wrapper_abc.TestWrapper):

    file = 'enterprise_pool_feed.txt'
    wrapper_class_to_test = ep.Pool

    def test_entries(self):
        self.assertEqual(1, len(self.entries))

    def test_type(self):
        self.assertIsInstance(self.dwrap, ep.Pool)

    def test_id(self):
        self.assertEqual(328, self.dwrap.id)

    def test_name(self):
        self.assertEqual('FVT_pool1', self.dwrap.name)

    def test_compliance_state(self):
        self.assertEqual(ep.ComplianceState.IN_COMPLIANCE,
                         self.dwrap.compliance_state)

    def test_compliance_hours_left(self):
        self.assertEqual(0, self.dwrap.compliance_hours_left)

    def test_total_mobile_procs(self):
        self.assertEqual(20, self.dwrap.total_mobile_procs)

    def test_total_mobile_mem(self):
        self.assertEqual(0, self.dwrap.total_mobile_mem)

    def test_avail_mobile_procs(self):
        self.assertEqual(16, self.dwrap.avail_mobile_procs)

    def test_avail_mobile_mem(self):
        self.assertEqual(0, self.dwrap.avail_mobile_mem)

    def test_unret_mobile_procs(self):
        self.assertEqual(0, self.dwrap.unret_mobile_procs)

    def test_unret_mobile_mem(self):
        self.assertEqual(0, self.dwrap.unret_mobile_mem)

    def test_mgmt_consoles(self):
        self.assertIsInstance(self.dwrap.mgmt_consoles,
                              entry_wrapper.WrapperElemList)
        self.assertEqual(1, len(self.dwrap.mgmt_consoles))
        console = self.dwrap.mgmt_consoles[0]
        self.assertIsInstance(console, ep.PoolMgmtConsole)
        self.assertEqual('ip9-1-2-3', console.name)
        self.assertEqual('7042-CR7*10B6EDC', console.mtms.mtms_str)
        self.assertTrue(console.is_master_console)
        self.assertEqual('9.1.2.3', console.ip_addr)

    def test_master_console_mtms(self):
        self.assertEqual('7042-CR7*10B6EDC',
                         self.dwrap.master_console_mtms.mtms_str)


class TestPoolMemberFile(test_wrapper_abc.TestWrapper):

    file = 'enterprise_pool_member_feed.txt'
    wrapper_class_to_test = ep.PoolMember

    def test_entries(self):
        self.assertEqual(5, len(self.entries))

    def test_type(self):
        for entry in self.entries:
            self.assertIsInstance(entry, ep.PoolMember)

    def test_mobile_procs(self):
        self.assertEqual(4, self.dwrap.mobile_procs)

    def test_mobile_mem(self):
        self.assertEqual(0, self.dwrap.mobile_mem)

    def test_set_mobile_procs(self):
        orig_value = self.dwrap.mobile_procs
        self.dwrap.mobile_procs = 999
        self.assertEqual(999, self.dwrap.mobile_procs)
        self.dwrap.mobile_procs = orig_value

    def test_set_mobile_mem(self):
        orig_value = self.dwrap.mobile_mem
        self.dwrap.mobile_mem = 888
        self.assertEqual(888, self.dwrap.mobile_mem)
        self.dwrap.mobile_mem = orig_value

    def test_inactive_procs(self):
        self.assertEqual(2, self.dwrap.inactive_procs)

    def test_inactive_mem(self):
        self.assertEqual(0, self.dwrap.inactive_mem)

    def test_unret_mobile_procs(self):
        self.assertEqual(0, self.dwrap.unret_mobile_procs)

    def test_unret_mobile_mem(self):
        self.assertEqual(0, self.dwrap.unret_mobile_mem)

    def test_proc_compliance_hours_left(self):
        self.assertEqual(0, self.dwrap.proc_compliance_hours_left)

    def test_mem_compliance_hours_left(self):
        self.assertEqual(0, self.dwrap.mem_compliance_hours_left)

    def test_sys_name(self):
        self.assertEqual('Server-8284-22A-SN21B63CV', self.dwrap.sys_name)

    def test_sys_installed_procs(self):
        self.assertEqual(20, self.dwrap.sys_installed_procs)

    def test_sys_installed_mem(self):
        self.assertEqual(524288, self.dwrap.sys_installed_mem)

    def test_sys_mtms(self):
        self.assertEqual('8284-22A*21B63CV', self.dwrap.sys_mtms.mtms_str)

    def test_sys_state(self):
        self.assertEqual('operating', self.dwrap.sys_state)

    def test_mgmt_consoles(self):
        self.assertIsInstance(self.dwrap.mgmt_consoles,
                              entry_wrapper.WrapperElemList)
        self.assertEqual(1, len(self.dwrap.mgmt_consoles))
        console = self.dwrap.mgmt_consoles[0]
        self.assertIsInstance(console, ep.PoolMgmtConsole)
        self.assertEqual('ip9-1-2-3', console.name)
        self.assertEqual('7042-CR7*10B6EDC', console.mtms.mtms_str)
        self.assertTrue(console.is_master_console)
        self.assertEqual('9.1.2.3', console.ip_addr)
